# -*- coding: utf-8 -*-
"""
@file
@brief Functions to prepare a setup on Windows
"""
from __future__ import print_function

import os
import shutil
import sys
import warnings
import datetime

from ..installhelper.install_cmd_helper import update_pip, run_cmd, python_version
from ..installhelper.module_dependencies import missing_dependencies

from .win_batch import create_win_batches
from .win_ipy_kernels import install_kernels
from .win_innosetup_helper import run_innosetup, innosetup_replacements
from .win_fix_compiler_c import switch_to_VS_compiler, switch_to_mingw_compiler
from .win_patch import win_patch_paths
from .win_setup_main_helper import dtnow, copy_icons, win_download, win_install, create_links_tools
from .win_setup_main_helper import win_download_notebooks, win_install_julia_step, win_install_r_step
from .win_packages import win_install_packages_other_python, get_modules_version
from .win_extract import clean_msi
from .win_ipython_helper import ipython_create_profile, ipython_update_profile, install_jupyter_extension
from .win_setup_r import get_package_description
from .win_exception import WinInstallMissingDependency
from .tutorial import copy_tutorial
from .win_setup_main_checkings import distribution_checkings

if sys.version_info[0] == 2:
    from codecs import open
    FileNotFoundError = Exception


license = """
Copyright (c) 2013-2016, Xavier Dupré

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

_default_notebooks = [
    ("docs/ensae", ["http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/_downloads/td1a_cenonce_session_12.ipynb",
                    "http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/_downloads/td2a_cenonce_session_2A.ipynb",
                    "http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/_downloads/td2a_cenonce_session_2B.ipynb",
                    "http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/_downloads/td2a_cenonce_session_2C.ipynb",
                    ]),
    ("docs/actuariat", ["http://www.xavierdupre.fr/app/actuariat_python/helpsphinx/_downloads/population_recuperation_donnees.ipynb",
                        ]),
]


def architecture():
    """
    @return either 32bit or 64bit
    """
    o, b = python_version()
    return b


def win_python_setup(folder="dist/win_python_setup_" + architecture(),
                     download_folder="build/win_python_setup_" + architecture(),
                     module_list=None, verbose=False, fLOG=print, download_only=False,
                     no_setup=False, notebooks=None, selection=None,
                     documentation=True, last_function=None, r_packages=True,
                     julia_packages=True, tutorial=None, source=None,
                     embed=True):
    """
    Prepares a Windows distribution of Python based on InnoSetup,
    inspired from WinPython but more easier to tweak I hope.

    @param      folder          where to prepare the python version
    @param      module_list     list of module to install (see @see fn small_installation = default options)
    @param      fLOG            logging function
    @param      download_only   only downloads
    @param      no_setup        skip the building of the setup
    @param      verbose         print more information
    @param      notebooks       notebooks to copy to the workspace, list of ("subfolders", url)
    @param      selection       selection of tools to install, example: ``{"R", "mingw", "tdm"}``,
                                empty by default
    @param      last_function   function to execute just before running InnoSetup,
                                see `win_setup_helper.py
                                <https://github.com/sdpython/ensae_teaching_cs/blob/master/src/
                                ensae_teaching_cs/automation/win_setup_helper.py>`_
                                for an example
    @param      r_packages      install R packages
    @param      julia_packages  install Julia packages
    @param      documentation   add documentation
    @param      tutorial        list of folders to copy in ``workspace/tutorial``,
                                it can refer to internal tutorials (see folder ``win_installer/tutorial``)
    @param      source          source of python packages (see @see cl ModuleInstall)
    @param      embed           custom *Python* or embedded distribution (False)
    @return                     list of completed operations

    The available tools to install must be chose among:

    * `R <http://www.r-project.org/>`_
    * `Julia <http://julialang.org/>`_
    * `MinGW <http://www.mingw.org/>`_
    * `TDM-GCC <http://tdm-gcc.tdragon.net/>`_
    * `VS <https://www.visualstudio.com/en-us/products/visual-studio-express-vs.aspx>`_
    * `Java JDK <http://www.oracle.com/technetwork/java/javase/downloads/jdk8-downloads-2133151.html>`_

    By default, only R is included. Julia requires too much work.
    The command line does not always end. The building of the package
    is sometimes reluctant to work. And the Julia kernel is exclusive:
    it cannot be setup with others kernel. Maybe the version 0.5 will fix those issues.

    The signature of function ``last_function`` should be the following::

        def last_function(inno_script, folders, verbose=False, fLOG=print):
            # something specific to do
            # before compiling the setup
            # such adding new tools
            # or replacing icons
            # the inno setup script is located in folders["logs"]

    The parameters *folders* is a dictionary which gives access to the main folders
    of the distribution.

    The setup will also contains `pandoc <http://pandoc.org/>`_,
    `7z <http://www.7-zip.org/>`_,
    `SQLiteSpy <https://www.yunqa.de/delphi/doku.php/products/sqlitespy/index>`_,
    `Scite <http://www.scintilla.org/SciTE.html>`_,
    `MinGW <http://www.mingw.org/>`_,
    `Graphviz <http://www.graphviz.org/>`_.

    The function first downloads everything.
    It does not do it twice, so you can run the function again and directly go
    to where it was interrupted. If there is no notebooks,
    the setup will add some anyway.

    It uses `InnoSetup <http://www.jrsoftware.org/isinfo.php>`_ to build the setup.

    The distribution will contain the following subfolders:

    * *tools*: subfolders for R, Julia, MinGW, Scite, pandoc, 7z, Putty...
    * *python*: subfolder for python interpreter
    * *workspace*: current folder for the notebooks
    * *build*: location of downloaded modules and tools

    Comments and remarks:

    * 7z setup needs a last click to complete
    * pandoc needs a last click to complete
    * R must be manually installed in the right folder
    * TDM-GCC is manually installed in the right folder
    * Julia produces a file exe, for the time being it must be done manually
    * MinGW is also installed manually, the command line is different from others tools,
      once it is installed, you should run the command line::

        mingw-get install binutils gcc g++ mingw32 fortran gdb mingw32 mingw w32api g77

    * Python setups needs a last click to complete
    * Julia command line sometimes gets stuck, the setup needs to be stopped
      and restarted. It happens while installing the packages and
      also while building IJulia (the package to use Julia in a notebook).
      The Julia should be stopped instead of stopping the python script.
      That trick shows the standard output of Julia.
    * Julia kernel cannot be used with the others: it requires a different
      configuration which prevents others kernel to be available at the same time.
      We will skip for the time being.
    * If the R kernel fails to start, you should manually run the script
      `R_install.r <https://github.com/sdpython/pymyinstall/blob/master/src/pymyinstall/win_installer/R_install.r>`_.

    With Julia, initialisation, installation or building takes time.
    The function writes a file ``log.step.<julia_step>.txt``
    to tell the step has completed once. You can remove this file
    to do it again.

    .. exref::
        :title: Prepare a standalone distribution

        The function downloads everything. The installation of tools
        is still manual. Package installation is automated.

        ::

            from pymyinstall import win_python_setup
            from pymyinstall.packaged import ensae_fullset
            list_modules = ensae_fullset()
            win_python_setup(module_list=list_modules,
                            verbose=False,
                            download_only=False)

        This works only for Windows.

    @warning The Julia installation might be stuck after the installation or the build.
             In that case, the script python should be stopped by *stopping the Julia
             process* from the Task Manager
             and started again. If it has completed, it will go to the
             next step.

    .. index:: issue

    **Known issues while preparing the setup:**

    * Some modules generates utf-8 encoding errors while being installed.
      The python scripts stops. It should be started again, it will
      detect the module was insalled and will go to the next one in the list.
    * The setup are started by the Python script but the user needs to manually
      click on the final ok button to proceed.

    **Known issues after the setup is installed:**

    * The first run of Spyder after the installation usually fails (failure of python.exe).
      The second one succeeds. You should run Spyder from the installation setup before
      compiling the setup.

    .. index:: missing modules, vcomp110.dll, llvmlite, numba, issue, theano, xgboost

    **Known extra steps needed by some modules**

    * **llvmlite**, **numba**, on Windows, if the dll
      *api-ms-win-crt-runtime-l1-1-0.dll* is missing, it is explained
      in `api-ms-win-crt-runtime-l1-1-0.dll error <https://github.com/cmderdev/cmder/issues/490>`_,
      `Visual C++ Redistributable for Visual Studio 2015 <https://www.microsoft.com/en-us/download/details.aspx?id=48145>`_
      needs to be installed.
    * **theano** requires `TDM-GCC <http://tdm-gcc.tdragon.net/>`_,
      read `Installation of Theano on Windows <http://deeplearning.net/software/theano/install_windows.html>`_
    * **xgboost** if DLL ``vcomp110.dll`` is missing, you should read blog :ref:`blog_xgboost_install`
      to understand how to get it.

    @todo Use chocolatey to process installation.

    @todo Fix Julia installation.
    """
    if selection is None:
        selection = set()
    if notebooks is None:
        notebooks = _default_notebooks

    if not isinstance(selection, set):
        selection = set(selection)
    selection.add("pandoc")
    selection.add("7z")
    selection.add("scite")
    selection.add("putty")
    selection.add("sqlitespy")
    selection.add("graphviz")
    selection.add("python")
    selection.add("jenkins")
    selection = set(_.lower() for _ in selection)

    _allowed = {"pandoc", "7z", "scite", "putty",
                "sqlitespy", "scite", "python", "tdm", "vs", "r", "graphviz",
                "jenkins", "jdk"}
    for s in selection:
        s_ = s.split("==")[0]
        if s_ not in _allowed:
            raise ValueError("{0} unknown, should be in {1}".format(
                s, ", ".join(sorted(_allowed))))

    fLOG("[pymy] --- selection", selection)

    #####
    # we change  for the version
    #####
    versions = {}
    for sel in selection:
        if "==" in sel:
            spl = sel.split("==")
            versions[spl[0].lower()] = spl[1]
        else:
            versions[sel.lower()] = None
    selection = versions

    ######
    # next
    ######

    operations = []
    operations.append(("time", dtnow()))

    folder = os.path.abspath(folder)
    download_folder = os.path.abspath(download_folder)

    if not os.path.exists(folder):
        os.makedirs(folder)
        operations.append(("mkdir", folder))

    if not os.path.exists(download_folder):
        os.makedirs(download_folder)
        operations.append(("mkdir", download_folder))

    ###################
    # definition of folders
    ###################
    operations.append(("time", dtnow()))
    folders = dict(tools=os.path.join(folder, "tools"),
                   workspace=os.path.join(folder, "workspace"),
                   python=os.path.join(folder, "python"),
                   config=os.path.join(folder, "config"),
                   logs=os.path.join(folder, "logs"),
                   build=download_folder,
                   )

    for k, v in folders.items():
        if not os.path.exists(v):
            os.mkdir(v)

    ###########################
    # download the documentation
    ###########################
    if documentation:
        operations.append(("documentation", "-"))
        op = win_download_notebooks(
            notebooks, folders["workspace"], verbose=verbose, fLOG=fLOG)
        operations.extend(op)
        operations.append(("time", dtnow()))

    ######################
    # download of everything
    ######################
    for mod in module_list:
        mod.fLOG = fLOG
    operations.append(("download", "-"))
    op = win_download(folder=download_folder,
                      module_list=module_list,
                      verbose=verbose, fLOG=fLOG,
                      selection=selection, source=source,
                      embed=embed)
    operations.extend(op)
    operations.append(("time", dtnow()))

    ########
    # license
    ########
    fLOG("[pymy] --- license")
    with open(os.path.join(folder, "license.txt"), "w") as f:
        f.write(license)
    operations.append(("license", "license.txt"))

    if not download_only:
        ########################
        # copy icons in tools/icons
        #######################
        fLOG("[pymy] --- copy icons")
        op = copy_icons(os.path.join(os.path.dirname(__file__), "icons"),
                        os.path.join(folders["tools"], "icons"))
        operations.extend(op)
        operations.append(("time", dtnow()))

        #############
        # install setups
        #############
        fLOG("[pymy] --- installation of python and tools")
        fLOG("[pymy] --- you might have to it yourself for R, Julia")
        op, installed = win_install(
            folders=folders, download_folder=download_folder, verbose=verbose, fLOG=fLOG,
            selection=selection)
        operations.extend(op)
        operations.append(("time", dtnow()))
        if "pandoc" not in installed:
            raise FileNotFoundError("pandoc was not installed")

        if verbose:
            for k, v in installed.items():
                fLOG("[pymy]   INSTALLED:", k, "-->", v)

        ##########
        # clean msi
        ##########
        fLOG("[pymy] --- clean msi")
        op = clean_msi(folders["tools"], "*.msi", verbose=verbose, fLOG=fLOG)
        operations.extend(op)
        operations.append(("time", dtnow()))

        ################
        # create links tools
        ################
        fLOG("[pymy] --- create links")
        op = create_links_tools(folder, installed, verbose=verbose, fLOG=fLOG)
        operations.extend(op)
        operations.append(("time", dtnow()))

        #########################
        # create batch command files
        #########################
        fLOG("[pymy] --- create batch command file")
        op = create_win_batches(
            folders, verbose=verbose, fLOG=fLOG, selection=selection, module_list=module_list)
        operations.extend(op)

        ###########
        # update pip
        ###########
        fLOG("[pymy] --- update pip")
        operations.append(("python pip", "update"))
        op = update_pip(folders["python"])
        operations.extend(op)
        operations.append(("time", dtnow()))

        if "julia" in selection and julia_packages:
            operations.append(("julia", "-"))
            ops = win_install_julia_step(folders, verbose=verbose, fLOG=fLOG)
            operations.extend(ops)
            operations.append(("time", dtnow()))

        if "r" in selection and r_packages:
            operations.append(("r", "-"))
            ops = win_install_r_step(folders, verbose=verbose, fLOG=fLOG)
            operations.extend(ops)
            operations.append(("time", dtnow()))

        ######################
        # installation of packages
        ######################
        fLOG("[pymy] --- installation of python packages")
        operations.append(("python packaes", "start"))
        python_path = folders["python"]
        win_install_packages_other_python(
            python_path, download_folder, verbose=verbose, fLOG=fLOG,
            module_list=module_list)
        fLOG("[pymy] done")
        operations.append(("time", dtnow()))

        ##########################
        # mingw, add file distutils.cfg
        ##########################
        if "mingw" in selection and "vs" not in selection:
            fLOG("[pymy] --- switch_to_mingw_compiler")
            op = switch_to_mingw_compiler(folders["python"])
            for o in op:
                operations.append(("modify", o))

        ##########################
        # Visual Studio, VS 2015 for Python 3.5
        ##########################
        if "vs" in selection:
            fLOG("[pymy] --- switch_to_VS_compiler")
            op = switch_to_VS_compiler(folders["python"])
            for o in op:
                operations.append(("modify", o))

        ######################
        # create jupyter profile
        ######################
        has_jupyter = False
        for mod in module_list:
            if mod.name == "jupyter":
                has_jupyter = True
        if has_jupyter:
            fLOG("[pymy] --- create jupyter profile")
            operations.append(("jupyter", "create profile"))
            ipath = ipython_create_profile(
                folders["config"], folders["python"], fLOG=fLOG)
            operations.append(("profile", ipath))
            operations.append(("time", dtnow()))

        ######################
        # update ipython profile
        ######################
        if has_jupyter:
            fLOG("[pymy] --- update jupyter profile")
            operations.append(("jupyter", "update profile"))
            ipython_update_profile(ipath)
            operations.append(("time", dtnow()))

        ######################
        # update jupyter extension
        ######################
        if has_jupyter:
            fLOG("[pymy] --- install jupyter extension")
            operations.append(("jupyter", "update install jupyter extension"))
            install_jupyter_extension(folders["python"])
            operations.append(("time", dtnow()))

        ######################
        # copy pywin32 dll to main folders
        ######################
        fLOG("[pymy] --- pywin32 dll to main folders")
        operations.append(("pywin32", "dll"))
        fdll = os.path.join(
            python_path, "Lib", "site-packages", "pywin32_system32")
        if not os.path.exists(fdll):
            fdll = os.path.join(
                python_path, "Lib", "site-packages", "pypiwin32_system32")
        if not os.path.exists(fdll):
            raise FileNotFoundError(fdll)
        for dll in os.listdir(fdll):
            full = os.path.join(fdll, dll)
            if os.path.isdir(full):
                continue
            try:
                shutil.copy(full, python_path)
            except KeyError as a:
                # it means it already exists
                continue
            operations.append(("pywin32", "copy " + dll))
        operations.append(("time", dtnow()))

        ########
        # kernels
        ########
        if has_jupyter:
            fLOG("[pymy] --- add kernels")
            operations.append(("kernel", "add"))
            res = install_kernels(folders["tools"], folders["python"])
            for r in res:
                fLOG("[pymy] ADD: kernel", r)
            operations.append(("time", dtnow()))

        #########
        # checking
        #########
        distribution_checkings(folders["python"], folders[
                               "tools"], fLOG=fLOG, module_list=module_list)

    ########
    # tutorial
    ########
    if tutorial is not None:
        fLOG("[pymy] --- copy tutorial")
        operations.append(("tutorial", "begin"))
        fold_tuto = os.path.join(folders["workspace"], "tutorial")
        if not os.path.exists(fold_tuto):
            fLOG("[pymy] --- create ", fold_tuto)
            operations.append(("create", fold_tuto))
            os.mkdir(fold_tuto)
        for tuto in tutorial:
            fLOG("[pymy] copy tutorial", tuto)
            operations.append(("tutorial", tuto))
            res = copy_tutorial(tuto, fold_tuto)
            for a, b, c in res:
                operations.append(("copy", c))

        operations.append(("time", dtnow()))

    ################################
    # prepare setup script for InnoSetup
    ###############################
    fLOG("[pymy] --- prepare setup script for InnoSetup")
    replacements = dict(__DISTPATH__=folder)
    new_script = innosetup_replacements(replacements=replacements, fLOG=fLOG,
                                        temp_folder=os.path.join(folders["logs"]))
    fLOG("[pymy] done")
    operations.append(("InnoSetup", "replacement"))
    operations.append(("time", dtnow()))

    if last_function is not None:
        #################
        # run last_function
        #################
        fLOG("[pymy] --- run last_function")
        operations.append(("start", "last_function"))
        last_function(new_script, folders, verbose=verbose, fLOG=fLOG)
        operations.append(("time", dtnow()))

    ##################
    # check there is no missing modules
    ##################
    if not download_only:
        scr = "from pymyinstall.installhelper import missing_dependencies;r=missing_dependencies();" + \
              "print('\\n'.join('\'{0}\' misses \'{1}\''.format(k,v) for k,v in sorted(r.items())))"
        cmd = '{0} -c "{1}"'.format(os.path.join(
            folders["python"], "python.exe"), scr)
        fLOG("[pymy] --- run dependencies")
        fLOG("[pymy] CMD:", cmd)
        out, err = run_cmd(cmd, wait=True)
        if len(err) > 0:
            raise WinInstallMissingDependency(err)
        fLOG(out)

    miss = missing_dependencies()
    if len(miss) > 0:
        mes = "\n".join("'{0}' misses '{1}'".format(k, ", ".join(v))
                        for k, v in sorted(miss.items()))
        warnings.warn(mes)

    ##################
    # patch exe in scripts
    ##################
    if not download_only:
        fLOG(
            "--- patch paths, see http://www.clemens-sielaff.com/create-a-portable-python-with-pip-on-windows/")
        op = win_patch_paths(
            os.path.join(folders["python"], "Scripts"), "", fLOG=fLOG)
        operations.extend(op)
        operations.append(("time", dtnow()))

    #################
    # print the list of modules (python)
    #################
    if not download_only:
        fLOG("[pymy] --- pip freeze")
        mods = get_modules_version(folders["python"])
        fLOG("[pymy] nb modules: {0}".format(len(mods)))
        if len(mods) == 0:
            raise ValueError(
                "unable to get module list from folder " + folders["python"])
        with open(os.path.join(folders["config"], "installed.python.packages.txt"), "w") as f:
            mods = [(a.lower(), a, b) for a, b in mods.items()]
            for la, a, b in sorted(mods):
                f.write("{0}\t{1}\n".format(a, b))

    #################
    # print the list of modules (R)
    #################
    if not download_only:
        r_path = os.path.join(folders["tools"], "R")
        r_lib = os.path.join(r_path, "library")
        if os.path.exists(r_lib):
            fLOG("[pymy] --- list R packages")
            packs = os.listdir(r_lib)
            with open(os.path.join(folders["config"], "installed.R.packages.txt"), "w") as f:
                packs_ = [(p.lower(), p) for p in packs]
                for lp, pack in sorted(packs_):
                    desc = get_package_description(r_path, pack)
                    vers = desc.get("Version", "unknown")
                    f.write("{0}\t{1}\n".format(pack, vers))

    if not no_setup:

        if not os.path.exists(folders["workspace"]):
            raise FileNotFoundError(folders["workspace"])
        if len(os.listdir(folders["workspace"])) == 0:
            raise FileNotFoundError(
                "folder {0} is empty, it should not".format(folders["workspace"]))

        # remove
        fLOG("[pymy] --- remove setup")
        dist = os.path.join(folders["logs"], "..", "dist", "setup")
        if os.path.exists(dist):
            exe = [_ for _ in os.listdir(dist) if ".exe" in _]
        else:
            exe = []
        if len(exe) > 0:
            for e in exe:
                os.remove(os.path.join(dist, e))

        ################################
        # prepare setup script for InnoSetup
        ###############################
        fLOG("[pymy] --- building setup with InnoSetup")
        out = run_innosetup(new_script, fLOG=fLOG,
                            temp_folder=os.path.join(folders["logs"]))
        with open(os.path.join(folders["logs"], "out.install.innosetup.txt"), "w", encoding="utf8") as f:
            f.write(out)
        fLOG("[pymy] done")
        operations.append(("InnoSetup", "done"))
        operations.append(("time", dtnow()))

        # copy
        fLOG("[pymy] --- copy setup")
        dist = os.path.join(folders["logs"], "..", "dist", "setup")
        to = os.path.join(folders["logs"], "..", "..")
        exe = [_ for _ in os.listdir(dist) if ".exe" in _]
        if len(exe) > 0:
            dt = datetime.datetime.now()
            suffix = "%d%02d%02d.%d" % (dt.year, dt.month, dt.day, dt.hour)
            for e in exe:
                shutil.copy(os.path.join(dist, e), to)
                operations.append(("copy", e))
                final = os.path.join(to, e)
                tof = final.replace(".exe", "_" + suffix + ".exe")
                operations.append(("rename", tof))
                os.rename(final, tof)

    ##########
    # store logs
    ##########
    with open(os.path.join(folders["logs"], "log.setup.txt"), "a", encoding="utf8") as f:
        f.write("\n")
        f.write("-------------------------------------------\n")
        f.write("NEW RUN\n")
        f.write("-------------------------------------------\n")
        for ab in operations:
            if isinstance(ab, tuple):
                if len(ab) == 2:
                    a, b = ab
                elif len(ab) == 1:
                    a, b = a, None
                else:
                    a, b = ab[0], str(ab[1:])

            if isinstance(b, str  # unicode#
                          ):
                b = b.replace(folder, "")
                b = b.replace(os.environ.get(
                    "USERNAME", os.environ["USER"]), "---")
            f.write("{0}\t{1}\n".format(a, b))



"""
@file
@brief Various function to install various python module from various location.
"""
from __future__ import print_function

import sys
import platform
import os
import zipfile
import datetime
from .module_install_exceptions import UpdatePipError
from .run_cmd import run_cmd_private, run_cmd_old

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def python_version():
    """
    Retrieves the platform and version of this :epkg:`python`.

    @return     tuple, example: ("win32","32bit") or ("win32","64bit")
    """
    return sys.platform, platform.architecture()[0]


def unzip_files(zipf, whereTo, fLOG=print):
    """
    Unzip files from a :epkg:`zip` archive.

    @param      zipf        archive
    @param      whereTo     destination folder
    @param      fLOG        logging function
    @return                 list of unzipped files
    """
    files = []
    with zipfile.ZipFile(zipf, "r") as file:
        for info in file.infolist():
            if not os.path.exists(info.filename):
                data = file.read(info.filename)
                tos = os.path.join(whereTo, info.filename)
                if not os.path.exists(tos):
                    finalfolder = os.path.split(tos)[0]
                    if not os.path.exists(finalfolder):
                        fLOG("    creating folder ", finalfolder)
                        os.makedirs(finalfolder)
                    if not info.filename.endswith("/"):
                        u = open(tos, "wb")
                        u.write(data)
                        u.close()
                        files.append(tos)
                        if fLOG:
                            fLOG("    unzipped ", info.filename, " to ", tos)
                elif not tos.endswith("/"):
                    files.append(tos)
            elif not info.filename.endswith("/"):
                files.append(info.filename)
    return files


def add_shortcut_to_desktop_for_module(name):
    """
    Adds a shortcut on a module which includes a script.

    @param      name        name of the module
    @return                 shortcut was added or not
    """
    if name == "spyder":
        from .link_shortcuts import add_shortcut_to_desktop, suffix
        from .module_install import ModuleInstall
        md = ModuleInstall("spyder", "exe", script="spyder.bat")
        sc = md.Script
        if os.path.exists(sc):
            ver = suffix()
            r = add_shortcut_to_desktop(sc, name + "." + ver, name + "." + ver)
            return os.path.exists(r)
        else:
            return False
    else:
        raise NotImplementedError(
            "nothing implemented for module: {0}".format(name))


def get_pip_program(exe=None):
    """
    Gets :epkg:`pip` executable and
    fixes an issue with :epkg:`Pandoc`.

    @param      exe             path to python executable
    @return                     pip executable

    .. faqref::
        :title: How can I check the dependencies?

        The module `pipdeptree <https://github.com/naiquevin/pipdeptree>`_ gives
        you something like::

            d3py==0.2.3
            - ipython [installed: 3.1.0]
            - networkx [installed: 1.9.1]
                - decorator [required: >=3.4.0, installed: 3.4.2]
            - numpy [installed: 1.9.2]
            - pandas [installed: 0.16.0]
                - pytz [required: >=2011k, installed: 2015.4]
                - python-dateutil [required: >=2, installed: 2.4.2]
                - six [required: >=1.5, installed: 1.9.0]
                - numpy [required: >=1.7.0, installed: 1.9.2]
            autopep8==1.1.1
            - pep8 [required: >=1.5.7, installed: 1.5.7]
            sphinxjp.themes.basicstrap==0.4.2
            - setuptools
            - Sphinx [installed: 1.3.1]
                - alabaster [required: >=0.7, installed: 0.7.4]
                - six [required: >=1.4, installed: 1.9.0]
                - colorama [installed: 0.3.3]
                - Pygments [required: >=2.0, installed: 2.0.2]
                - Babel [required: >=1.3, installed: 1.3]
                - pytz [required: >=0a, installed: 2015.4]
                - snowballstemmer [required: >=1.1, installed: 1.2.0]
                - docutils [required: >=0.11, installed: 0.12]
                - sphinx-rtd-theme [required: >=0.1, installed: 0.1.8]
                - Sphinx [required: >=1.3, installed: 1.3.1]
                    - alabaster [required: >=0.7, installed: 0.7.4]
                    - six [required: >=1.4, installed: 1.9.0]
                    - colorama [installed: 0.3.3]
                    - Pygments [required: >=2.0, installed: 2.0.2]
                    - Babel [required: >=1.3, installed: 1.3]
                    - pytz [required: >=0a, installed: 2015.4]
                    - snowballstemmer [required: >=1.1, installed: 1.2.0]
                    - docutils [required: >=0.11, installed: 0.12]
                    - Jinja2 [required: >=2.3, installed: 2.7.3]
                    - MarkupSafe [installed: 0.23]
                - Jinja2 [required: >=2.3, installed: 2.7.3]
                - MarkupSafe [installed: 0.23]

            ...
    """
    tried = []
    if exe is None:
        exe = os.path.dirname(sys.executable)
    major, minor = sys.version_info[0:2]
    if sys.platform.startswith("win"):
        if not exe.lower().endswith("scripts"):
            pi = os.path.join(exe, "Scripts", "pip.exe")
            tried.append(pi)
            if not os.path.exists(pi):
                pi = os.path.join(exe, "Scripts", "pip%d.exe" % major)
                tried.append(pi)
                if not os.path.exists(pi):
                    # Anaconda is different
                    pi = os.path.join(exe, "Scripts", "pip.exe")
                    tried.append(pi)
                    if not os.path.exists(pi):
                        pi = os.path.join(exe, "Scripts", "pip%d.exe" % major)
                        tried.append(pi)
                        if not os.path.exists(pi):
                            pi = os.path.join(
                                exe, "Scripts", "pip%d.%d.exe" % (major, minor))
                            tried.append(pi)
                            raise FileNotFoundError(
                                "tried (1):\n" + "\n".join(tried) + "\n---- try ---\npython -m pip install -U pip --force")
        else:
            pi = os.path.join(exe, "pip.exe")
            tried.append(pi)
            if not os.path.exists(pi):
                # Anaconda is different
                pi = os.path.join(exe, "pip.exe")
                tried.append(pi)
                if not os.path.exists(pi):
                    pi = os.path.join(exe, "pip%d.exe" % major)
                    tried.append(pi)
                    if not os.path.exists(pi):
                        pi = os.path.join(exe, "pip%d.%d.exe" % (major, minor))
                        tried.append(pi)
                        if not os.path.exists(pi):
                            raise FileNotFoundError(
                                "tried (2):\n" + "\n".join(tried) + "\n---- try ---\npython -m pip install -U pip --force")
    else:
        if sys.version_info[0] == 2:
            if exe is None:
                return "pip"
            else:
                pi = os.path.join(exe, "pip")
        else:
            major = sys.version_info[0]
            minor = sys.version_info[1]
            if exe is None:
                return "pip%d.%d" % (major, minor)
            else:
                # this does not work because on Linux, the binary is installed on the local path
                # pip3.4 are not in the same place
                # pi = os.path.join(exe, "pip%d.%d" % (major, minor))
                import pip
                exe = os.path.normpath(os.path.join(os.path.dirname(
                    pip.__file__), "..", "..", "..", "..", "bin"))
                pi = os.path.join(exe, "pip%d.%d" % (major, minor))
                if not os.path.exists(pi):
                    pi = os.path.join(exe, "pip")
                    if not os.path.exists(pi):
                        raise FileNotFoundError(
                            "unable to find pip: {0}\n__file__={1}\nexe={2}".format(pi, pip.__file__, exe))
    return pi


def get_python_program():
    """
    Returns the executable for :epkg:`python`.

    .. versionadded:: 1.1
    """
    pip = get_pip_program()
    dirname = os.path.dirname(pip)
    exe = os.path.join(
        dirname, "python.exe" if sys.platform.startswith("win") else "python")
    if os.path.exists(exe):
        return exe
    exe = os.path.normpath(os.path.join(
        dirname, "..", "python.exe" if sys.platform.startswith("win") else "python"))
    if os.path.exists(exe):
        return exe
    raise FileNotFoundError(exe)


def get_conda_program(exe=None):
    """
    Gets :epkg:`conda` executable and
    fixes an issue with :epkg:`Pandoc`.

    @param      exe             path to python executable
    @return                     conda executable
    """
    tried = []
    if exe is None:
        exe = os.path.dirname(sys.executable)
    if sys.platform.startswith("win"):
        if not exe.lower().endswith("scripts"):
            pi = os.path.join(exe, "Scripts", "conda.exe")
            tried.append(pi)
            if not os.path.exists(pi):
                # Anaconda is different
                pi = os.path.join(exe, "Scripts", "conda.exe")
                tried.append(pi)
                if not os.path.exists(pi):
                    raise FileNotFoundError(
                        "tried (1):\n" + "\n".join(tried))
        else:
            pi = os.path.join(exe, "conda.exe")
            tried.append(pi)
            if not os.path.exists(pi):
                # Anaconda is different
                pi = os.path.join(exe, "conda.exe")
                tried.append(pi)
                if not os.path.exists(pi):
                    raise FileNotFoundError(
                        "tried (2):\n" + "\n".join(tried))
    else:
        if exe is None:
            return "conda"
        else:
            pi = os.path.join(exe, "conda")

    return pi


def get_file_modification_date(filename):
    """
    Gets the date modification for a filename.

    @param      filename        filename
    @return                     datetime
    """
    t = os.path.getmtime(filename)
    return datetime.datetime.fromtimestamp(t)


def update_pip(python_path=None, fLOG=print):
    """
    Updates :epkg:`pip` for a specific distribution.

    @param      python_path     python path (or sys.executable if None)
    @param      fLOG            logging function
    @return                     output

    The command ``python -m pip install -U pip`` or
    ``pip install --upgrade pip`` might fail on Windows due to very long paths
    (see `Upgrading pip fails on Windows when install path is too long <https://github.com/pypa/pip/issues/3055>`_).
    If that happens,
    assuming the module *pymyinstall* was installed with pip, we can now remove
    *pip* and use *get_pip.py* instead. This part requires *pyquickhelper*.

    We try the url `bootstrap.pypa.io/get-pip.py <https://bootstrap.pypa.io/get-pip.py>`_ first
    then a local copy.
    """
    if python_path is None:
        python_path = sys.executable
    else:
        python_path = os.path.join(python_path, "python")
    cmd = python_path + " -m pip install -U pip"
    out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
    if err and len(err) > 0:
        if ("FileNotFoundError" in err or "No module named pip.__main__" in err) \
           and sys.platform.startswith("win"):
            from pyquickhelper.filehelper import remove_folder
            # we try to remove pip and to install it again
            # it might be due to long path on Windows
            pack = os.path.join(os.path.dirname(
                python_path), "Lib", "site-packages")
            if not os.path.exists(pack):
                raise FileNotFoundError(pack)
            fpip = os.path.join(pack, "pip")
            if os.path.exists(fpip):
                # remove the folder
                fLOG("  remove folder", fpip)
                remove_folder(fpip)

            pip_ = [_ for _ in os.listdir(pack) if _.startswith("pip-")]
            if len(pip_) > 0:
                for _ in pip_:
                    fp = os.path.join(pack, _)
                    fLOG("  remove folder", fp)
                    remove_folder(fpip)

            url = "https://bootstrap.pypa.io/get-pip.py"
            cmd = python_path + " " + url
            out, err = run_cmd(cmd, wait=True)
            if err and len(err) > 0:
                get_pip = os.path.abspath(os.path.join(
                    os.path.dirname(__file__), "get_pip.py"))
                if not os.path.exists(get_pip):
                    raise FileNotFoundError(get_pip)
                cmd = python_path + " " + get_pip
                out, err = run_cmd(cmd, wait=True)
                if err and len(err) > 0:
                    raise UpdatePipError(
                        "unable to update pip with get_pip.\nCMD:\n{0}\nOUT:\n{1}\nERR-E:\n{2}".format(cmd, out, err))
        else:
            lines = err.split("\n")
            keep = []
            for line in lines:
                if len(line.strip("\n\r\t ")) == 0:
                    continue
                if "Prompt dismissed.." in line:
                    continue
                if not line.startswith(" ") and "RuntimeWarning: Config variable" not in line and \
                    not(" which is incompatible." in line and " has requirement " in line) and \
                        not(" requires " in line and " which is not installed." in line) and \
                        not("Cache entry deserialization failed, entry ignored" in line) and \
                        len(line.strip()) > 3:
                    keep.append(line)
            if len(keep) > 0 and "Requirement already up-to-date" not in out:
                for _ in keep:
                    print("++", _)
                raise UpdatePipError(
                    "Unable to update pip.\nCMD:\n{0}\nOUT:\n{1}\nERR-F:"
                    "\n{2}\n---KEPT---\n{3}".format(cmd, out, err, "\n".join(keep)))
    return out


def has_pip():
    """
    Tells if :epkg:`pip` is installed.

    @return     boolean
    """
    try:
        import pip
        return pip is not None
    except ImportError:
        return False


def is_conda_distribution():
    """
    Tells if it is a :epkg:`conda` distribution or not,
    check the presence of ``Continuum Analytics``
    or ``|Anaconda`` in ``sys.version``.

    @return         boolean

    .. versionadded:: 1.1
    """
    return "Continuum Analytics" in sys.version or "|Anaconda" in sys.version


def run_cmd(cmd, sin="", shell=sys.platform.startswith("win"), wait=False, log_error=True,
            stop_running_if=None, encerror="ignore",
            encoding="utf8", change_path=None, communicate=True,
            preprocess=True, timeout=None, catch_exit=False, fLOG=None,
            tell_if_no_output=None, old_behavior=False):
    """
    Runs a command line and waits for the results,
    @see fn run_cmd_private.
    """
    if old_behavior or not sys.platform.startswith("win"):
        return run_cmd_old(
            cmd=cmd, sin=sin, shell=shell, wait=wait, log_error=log_error,
            secure=None, stop_waiting_if=stop_running_if, do_not_log=False,
            encerror=encerror, encoding=encoding, cwd=change_path, fLOG=fLOG)
    return run_cmd_private(
        cmd=cmd, sin=sin, shell=shell, wait=wait, log_error=log_error,
        stop_running_if=stop_running_if, encerror=encerror,
        encoding=encoding, change_path=change_path, communicate=communicate,
        preprocess=preprocess, timeout=timeout, catch_exit=catch_exit, fLOG=fLOG,
        tell_if_no_output=tell_if_no_output)



"""
@file
@brief Change Jupyter configuration
"""
import sys
import os
import re

from ..installhelper.link_shortcuts import add_shortcut_to_desktop, suffix


def noLOG(*args, **kwargs):
    pass


def install_mathjax():
    """
    install a local copy of mathjax
    """
    from IPython.external import mathjax  # pylint: disable=E1101,E0611
    mathjax.install_mathjax()


def setup_ipython(current_path=None, additional_path=None, apply_modification=True,
                  shortcut=True, browser=None):
    """
    The function applies the modification suggested in this blog post:
    `Travailler avec IPython notebook <http://www.xavierdupre.fr/blog/2014-02-24_nojs.html>`_ (written in French).

    @param      additional_path     additional paths to add to jupyter (a list)
    @param      current_path        change the current path when running a notebook
    @param      apply_modification  apply the modification, otherwise, just create the profile
    @param      shortcut            add short cut the desktop
    @param      browser             if not None, tries to change if it finds it (IE, Chrome, Firefox, on Windows)
    @return                         path the config file

    If you need to create a shortcut with the appropriate paths,
    you can use the following instructions
    to open :epkg:`Jupyter` notebook on a specific folder:

    ::

        set path=%path%;c:\\Python370_x64;c:\\Python370_x64\\Scripts
        jupyter-notebook --notebook-dir=_doc\\notebooks
    """
    if sys.platform.startswith("win"):
        user_profile = os.environ['USERPROFILE']
        profile = os.path.join(user_profile, ".jupyter")
        ipython_notebook_config = os.path.join(
            profile,
            "jupyter_notebook_config.py")

        checkpath = os.path.join(profile, "checkspoints")
        if not os.path.exists(checkpath):
            os.mkdir(checkpath)

        if not os.path.exists(ipython_notebook_config):
            from ..installhelper.install_cmd_helper import run_cmd
            exe_exe = os.path.dirname(sys.executable)
            if exe_exe.lower().endswith("scripts"):
                exe = os.path.join(exe_exe, "jupyter-notebook.exe")
            else:
                exe = os.path.join(exe_exe, "Scripts", "jupyter-notebook.exe")
            cmd = exe + " -y --generate-config"
            out, err = run_cmd(cmd, wait=True, fLOG=noLOG)

            if not os.path.exists(ipython_notebook_config):
                raise Exception(
                    "unable to create jupyter configuration in \n'{0}'\nbecause of:\n{1}\nERR-4:\n{2}\ncmd={3}".format(
                        ipython_notebook_config, out, err, cmd))

        with open(ipython_notebook_config, "r") as f:
            text = f.read()

        # change current path and pylab configuration
        for var in ["IPKernelApp.file_to_run",
                    "ProfileDir.location",
                    "FileNotebookManager.checkpoint_dir",
                    "NotebookManager.notebook_dir",
                    "NotebookApp.ipython_dir",
                    "IPKernelApp.pylab"]:
            reg = re.compile("(#? *c.{0} =.*)".format(var))
            alls = reg.findall(text)
            if len(alls) == 1 and current_path is not None:
                if "pylab" in var:
                    text = text.replace(
                        alls[0],
                        "c.{0} = 'inline'".format(var))
                elif "checkpoint_dir" in var:
                    text = text.replace(
                        alls[0],
                        "c.{0} = r'{1}'".format(
                            var,
                            checkpath))
                elif "file_to_run" not in var:
                    text = text.replace(
                        alls[0],
                        "c.{1} = r'{0}'".format(
                            current_path,
                            var))
                else:
                    text = text.replace(
                        alls[0],
                        "c.{1} = '{0}\\jupyter_startup.py'".format(
                            current_path,
                            var))

        # browser
        if browser is not None:
            if sys.platform.startswith("win"):
                paths = {"firefox": "C:\\Program Files (x86)\\Mozilla Firefox\\firefox.exe",
                         "ie": "C:\\Program Files\\Internet Explorer\\iexplore.exe",
                         "chrome": "C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe",
                         }
                browser = browser.lower()
                if browser not in paths:
                    raise KeyError(
                        "unable to find browser {0} in [{1}]".format(
                            browser, ",".join(
                                paths.keys())))
                subtext = """
                    import webbrowser
                    webbrowser.register('{0}', None, webbrowser.GenericBrowser(r'{1}'))
                    c.NotebookApp.browser = '{0}'
                    """.format(browser, paths[browser]).replace("                    ", "")
                text += subtext

            else:
                raise NotImplementedError("implemented only on Windows")

        if apply_modification:
            # write modification
            with open(ipython_notebook_config, "w") as f:
                f.write(text)

        # write jupyter_startup.py
        rows = ["import sys"]
        if additional_path is not None:
            for p in additional_path:
                if not os.path.exists(p):
                    raise FileNotFoundError(p)
                rows.append("sys.path.append(r'{0}')".format(p))
        s = "\n".join(rows)

        if apply_modification:
            with open(os.path.join(current_path, "jupyter_startup.py"), "w") as f:
                f.write(s)

        return [ipython_notebook_config]

    else:
        raise NotImplementedError("implemented only for Windows")


def add_shortcut_to_desktop_for_ipython(folder):
    """
    create a shortcut on your desktop

    @param      folder      notebook dir
    @return                 filename
    """
    file = os.path.join(
        os.path.split(
            sys.executable)[0],
        "Scripts",
        "jupyter-notebook")
    arguments = " --notebook-dir=" + folder
    ver = suffix()
    return add_shortcut_to_desktop(
        file, "notebook." + ver, "Jupyter Notebook {1} ({0})".format(folder, ver), arguments)


if __name__ == "__main__":
    setup_ipython("C:\\temp", [], apply_modification=False)



# -*- coding: utf-8 -*-
"""
@file
@brief Various functions to install `python <http://www.python.org/>`_.
"""
from __future__ import print_function
import sys
import os
import datetime

from ..installhelper.install_cmd_helper import run_cmd, unzip_files
from .install_custom import download_page, download_file

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def unzip7_files(filename_7z, fLOG=print, dest="."):
    """
    If `7z <http://www.7-zip.org/>`_ is installed, the function uses it
    to uncompress file into *7z* format. The file *filename_7z* must not exist.

    .. index:: 7zip, 7z

    :param      filename_7z:     final destination
    :param      fLOG:            logging function
    :param      dest:            destination folder

    :return:                     output of 7z

    .. versionadded:: 1.1
    """
    if sys.platform.startswith("win"):
        exe = r"C:\Program Files\7-Zip\7z.exe"
        if not os.path.exists(exe):
            raise FileNotFoundError("unable to find: {0}".format(exe))
    else:
        exe = "7z"

    if not os.path.exists(filename_7z):
        raise FileNotFoundError(filename_7z)

    cmd = '"{0}"-y -o"{2}" x "{1}"'.format(exe, filename_7z, dest)
    out, err = run_cmd(cmd, wait=True)

    if err is not None and len(err) > 0:
        raise Exception("OUT:\n{0}\nERR-A:\n{1}".format(out, err))

    return out


def fix_fcntl_windows(path):
    """
    Adds a file `fnctl.py` on :epkg:`Windows`
    (only available on :epkg:`Linux`).

    @param   path       path to the python installation
    """
    if not sys.platform.startswith("win"):
        raise Exception("fcntl should only be added on Windows.")
    dest = os.path.join(path, "Lib", "fcntl.py")
    if os.path.exists(dest):
        # already done
        return
    module = """
                def fcntl(fd, op, arg=0):
                    return 0
                def ioctl(fd, op, arg=0, mutable_flag=True):
                    if mutable_flag:
                        return 0
                    else:
                        return ""
                def flock(fd, op):
                    return
                def lockf(fd, operation, length=0, start=0, whence=0):
                    return
        """.replace("                ", "")
    with open(dest, "w") as f:
        f.write(module)


def fix_termios_windows(path):
    """
    Adds a file `termios.py` on :epkg:`Windows`
    (only available on :epkg:`Linux`).

    @param   path       path to the python installation
    """
    if not sys.platform.startswith("win"):
        raise Exception("fcntl should only be added on Windows.")
    dest = os.path.join(path, "Lib", "termios.py")
    if os.path.exists(dest):
        # already done
        return
    module = """
                TCSAFLUSH = 1
        """.replace("                ", "")
    with open(dest, "w") as f:
        f.write(module)


def fix_resource_windows(path):
    """
    Adds a file `resource.py` on :epkg:`Windows`
    (only available on :epkg:`Linux`).

    @param   path       path to the python installation
    """
    if not sys.platform.startswith("win"):
        raise Exception("fcntl should only be added on Windows.")
    dest = os.path.join(path, "Lib", "resource.py")
    if os.path.exists(dest):
        # already done
        return
    module = """
        """.replace("                ", "")
    with open(dest, "w") as f:
        f.write(module)


def _clean_err1(err):
    if err:
        lines = []
        for line in err.split("\n"):
            if "find: ‘build’: No such file or directory" in line:
                continue
            if "(ignored)" in line:
                continue
            if "Task was destroyed but it is pending!" in line:
                continue
            if "[libinstall] Error 1 (ignored)" in line:
                continue
            if "task: <Task finished coro=<<async_generator_athrow without __name__>()" in line:
                continue
            if "stty: 'standard input': Inappropriate ioctl for device" in line:
                continue
            if "task: <Task pending coro=<<async_generator_athrow without __name__>()>>" in line:
                continue
            if "unhandled exception during asyncio.run() shutdown" in line:
                continue
            if "RuntimeError: can't send non-None value to a just-started coroutine" in line:
                continue
            if " which is not installed." in line:
                continue
            lines.append(line)
        err = "\n".join(lines).strip() if lines else None
        errl = err.lower()
        if 'error' not in errl and 'exception' not in errl:
            lines = []
            for line in err.split("\n"):
                if line.startswith('  '):
                    continue
                if 'note: declared here' in line:
                    continue
                if "In file included" in line:
                    continue
                if "warning:" in line:
                    continue
                if "In function " in line:
                    continue
                lines.append(line)
        err = "\n".join(lines).strip() if lines else None
    return err


def _clean_err0(err):
    # remove a couple of warnings.
    lines = err.split("\n")
    lines2 = [
        _ for _ in lines if "UserWarning: Module pymyinstall was already imported" not in _]
    if len(lines2) < len(lines):
        lines2 = [
            _ for _ in lines2 if "from pip._vendor import pkg_resources" not in _]
    return "\n".join(lines2)


def install_python(temp_folder=".", fLOG=print, install=True, force_download=False,  # pylint: disable=R0914
                   version=None, modules=None, custom=False, latest=False,
                   download_folder="download", verbose=False, make_first=False):
    """
    Installs :epkg:`python`.
    It does not do it a second time if it is already installed.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of python
    @param      version         version to download (by default the current version of Python)
    @param      modules         modules to install
    @param      custom          the standalone distribution has issue when installing new packages,
                                custom is True means switching to a zip of the standard distribution,
                                see below
    @param      latest          install this version of pymyinstall and not the pypi version
    @param      download_folder download folder for packages
    @param      verbose         more display
    @param      make_first      run *make* before *make altinstall*
    @return                     temporary file

    The version is fixed to the current version of Python and amd64.
    The standalone distribution has an issue and raises an error for some
    packages such as `smart_open <https://pypi.python.org/pypi/smart_open>`_:

    ::

        error: [Errno 2] No such file or directory: '<python>\\python36.zip\\lib2to3\\Grammar.txt'

    In that case, you should consider using ``custom=True``.
    The function work for :epkg:`Linux` too.
    List of steps done in linux:

    ::

        mkdir install_folder
        cd install_folder
        curl -O https://www.python.org/ftp/python/3.7.2/Python-3.7.2.tgz
        tar xzf Python-3.7.2.tgz
        mkdir dist372
        cd Python-3.7.2/
        # current folder is /home/dupre/temp/temp_py/dist372/
        ./configure --enable-optimizations --with-ensurepip=install --prefix=/home/dupre/temp/temp_py/dist372/inst --exec-prefix=/home/dupre/temp/temp_py/dist372/bin --datadir=/home/dupre/temp/temp_py/dist372/data
    """
    cmds = []

    if version is None:
        version = "%s.%s.%s" % sys.version_info[:3]
    versioni = tuple(int(_)
                     for _ in version.split("."))  # pylint: disable=R1728
    link = "https://www.python.org/downloads/release/python-%s/" % version.replace(
        ".", "")
    page = download_page(link)
    if page is None:
        raise ValueError("page is None for link '{0}'".format(link))

    if sys.platform.startswith("win"):
        if versioni[:2] <= (3, 4):
            raise NotImplementedError(
                "Python <= 3.4 is not supported anymore.")
        # The setup for Python 3.5 does not accept multiple versions,
        # it was installed on one machine and then compressed into a 7z
        # file
        if versioni >= (3, 7, 0):
            if custom:
                if versioni > (3, 7, 0):
                    raise ValueError(
                        "Not custom zip available for Python {0}".format(versioni))
                url = "http://www.xavierdupre.fr/enseignement/setup/Python{0}{1}-{0}.{1}.{2}-amd64.zip".format(
                    *versioni[:3])
            else:
                url = "https://www.python.org/ftp/python/{0}.{1}.{2}/python-{0}.{1}.{2}-embed-amd64.zip".format(
                    *versioni[:3])
        elif versioni >= (3, 6, 0):
            if custom:
                if versioni > (3, 6, 5):
                    raise ValueError(
                        "Not custom zip available for Python {0}".format(versioni))
                url = "http://www.xavierdupre.fr/enseignement/setup/Python{0}{1}-{0}.{1}.{2}-amd64.zip".format(
                    *versioni[:3])
            else:
                url = "https://www.python.org/ftp/python/{0}.{1}.{2}/python-{0}.{1}.{2}-embed-amd64.zip".format(
                    *versioni[:3])
        elif versioni >= (3, 5, 0):
            if custom:
                if versioni not in [(3, 5, 3), (3, 5, 2)]:
                    raise ValueError(
                        "Not custom zip available for Python {0}".format(versioni))
                url = "http://www.xavierdupre.fr/enseignement/setup/Python35-3.5.3-amd64.zip"
            else:
                url = "https://www.python.org/ftp/python/3.5.3/python-3.5.3-embed-amd64.zip"
        else:
            raise Exception(
                "Unable to find a proper version for version {0}".format(version))
    else:
        url = "https://www.python.org/ftp/python/{0}.{1}.{2}/Python-{0}.{1}.{2}.tgz".format(
            *versioni)

    full = url.split("/")[-1]
    outfile = os.path.join(temp_folder, full)
    fLOG("[install_python] download", url)
    local = download_file(url, outfile, fLOG=fLOG)

    # Install
    if install:
        # unzip files
        if sys.platform.startswith("win"):
            unzip_files(local, temp_folder, fLOG=fLOG)
        else:
            cmd = "tar xzf {0}".format(outfile)
            out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                               change_path=temp_folder)
            cmds.append(cmd)
            if err:
                raise RuntimeError(
                    "Issue with running '{0}'\n--OUT--\n{1}\n--ERR--\n{2}\n--IN--\n{3}\n--CMDS--\n{4}".format(
                        cmd, out, err, temp_folder, "\n".join(cmds)))
            pyinstall = os.path.join(
                temp_folder, "Python-{0}.{1}.{2}".format(*versioni))

            cmd = "./configure --enable-optimizations --with-ensurepip=install --prefix={0}/inst --exec-prefix={0}/bin --datadir={0}/data"
            cmd = cmd.format(temp_folder)
            out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                               change_path=pyinstall)
            cmds.append(cmd)
            if err:
                lines = []
                for line in err.split("\n"):
                    if "[libinstall] Error 1 (ignored)" in line:
                        continue
                    lines.append(line)
                err = "\n".join(lines).strip() if lines else None
            if err:
                raise RuntimeError(
                    "Issue with running '{0}'\n--OUT--\n{1}\n--ERR--\n{2}\n--CMDS--\n{3}".format(
                        cmd, out, err, "\n".join(cmds)))

            # See https://stackoverflow.com/questions/44708262/make-install-from-source-python-without-running-tests.
            os.environ["EXTRATESTOPTS"] = "--list-tests"

            if make_first:
                cmd = "make"
                out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                                   change_path=pyinstall)
                cmds.append(cmd)
                err = _clean_err1(err)
                if err:
                    raise RuntimeError(
                        "Issue while running '{0}'\n---URL---\n{1}\n---OUT---\n{2}\n"
                        "---ERR---?1-\n{3}\n---IN---\n{4}\n---CMDS---\n{5}".format(
                            cmd, url, out, err, pyinstall, "\n".join(cmds)))

            cmd = "make altinstall"
            out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                               change_path=pyinstall)
            cmds.append(cmd)
            err = _clean_err1(err)
            if err:
                lines = []
                for line in err.split("\n"):
                    if "[libinstall] Error 1 (ignored)" in line:
                        continue
                    if ' which is not on PATH.' in line:
                        continue
                    lines.append(line)
                err = "\n".join(lines).strip() if lines else None
            if err:
                raise RuntimeError(
                    "Issue while running '{0}'\n---URL---\n{1}\n---OUT---\n{2}\n---ERR---?2-\n{3}\n---IN---\n{4}\n---CMDS---\n{5}".format(
                        cmd, url, out, err, pyinstall, "\n".join(cmds)))

        # has pip?
        if sys.platform.startswith("win"):
            pyexe = os.path.join(temp_folder, "python.exe")
        else:
            pyexe = os.path.join(temp_folder, "bin", "python")
        cmd = "{0} -m pip --help"
        cmds.append(cmd)
        try:
            _, err = run_cmd(cmd, wait=True)
            has_pip = not err
        except Exception:
            has_pip = False

        # get-pip
        if not has_pip:
            get_pip = "https://bootstrap.pypa.io/get-pip.py"
            outfile_pip = os.path.join(temp_folder, "get-pip.py")
            download_file(get_pip, outfile_pip, fLOG=fLOG)

            # following issue https://github.com/pypa/get-pip/issues/7
            if sys.platform.startswith("win"):
                vers = "%d%d" % versioni[:2]
                if vers in ("36", "37"):
                    pth = os.path.join(temp_folder, "python%s._pth" % vers)
                    if os.path.exists(pth):
                        with open(pth, "r") as f:
                            content = f.read()
                        content = content.replace(
                            "#import site", "import site")
                        with open(pth, "w") as f:
                            f.write(content)

        # run get-pip.py
        if sys.platform.startswith("win"):
            pyexe = os.path.join(temp_folder, "python.exe")
        else:
            versioni3 = versioni[:3]
            pyexe = os.path.join(
                temp_folder, "Python-{}.{}.{}".format(*versioni3), "python")
        if not os.path.exists(pyexe):
            raise FileNotFoundError(pyexe)

    # Patches for windows.
    if install and sys.platform.startswith("win"):
        if not custom:
            cmd = '"{0}" -u "{1}"'.format(pyexe, outfile_pip)
            out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
            cmds.append(cmd)
            if len(err) > 0:
                skip = ['Consider adding this directory to PATH',
                        'which is not on PATH.']
                lines = err.split('\n')
                errs = []
                for line in lines:
                    zoo = True
                    for sk in skip:
                        if sk in line:
                            zoo = False
                            break
                    if zoo:
                        errs.append(line)
                err = "\n".join(errs).strip(' \n\r')
            if len(err) > 0:
                raise Exception(
                    "Something went wrong:\nCMD\n{0}\nOUT\n{1}\nERR-B\n{2}\n---CMDS--\n{3}".format(
                        cmd, out, err, "\n".join(cmds)))
        else:
            from ..win_installer.win_patch import win_patch_paths
            fLOG("[install_python] Patch scripts .exe")
            patched = win_patch_paths(temp_folder, pyexe, fLOG=fLOG)
            for pat in patched:
                fLOG("  - ", pat)

        # fix fcntl
        fix_fcntl_windows(temp_folder)
        fix_termios_windows(temp_folder)
        fix_resource_windows(temp_folder)

    # modules
    if install and modules is not None:
        if isinstance(modules, list):
            raise NotImplementedError(
                "Not implemented for a list of modules.")

        # cmd = '"{0}" -u -c "import pip;pip.main([\'install\',
        #        \'https://github.com/sdpython/pymyinstall/archive/master.zip\'])"'.format(pyexe)
        cmd = '"{0}" -u -c "import pip._internal;pip._internal.main([\'install\', \'pyquicksetup\'])"'.format(
            pyexe)
        fLOG("[install_python] " + cmd)
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG, change_path=None)
        cmds.append(cmd)
        if latest:
            folder = os.path.normpath(os.path.join(os.path.abspath(
                os.path.dirname(__file__)), "..", "..", ".."))
            setup = os.path.join(folder, "setup.py")
            if not os.path.exists(setup):
                raise FileNotFoundError(setup)
            sep = "\\" if sys.platform.startswith("win") else "/"
            cmd = '"{0}" -u "{1}{2}setup.py" install'.format(
                pyexe, folder, sep)
            change_path = folder
        else:
            cmd = '"{0}" -u -c "import pip._internal;pip._internal.main([\'install\', \'pymyinstall\'])"'.format(
                pyexe)
            change_path = None
        fLOG("[install_python] " + cmd)
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG, change_path=change_path)
        cmds.append(cmd)
        err_keep = err
        err = [_ for _ in err.split("\n")
               if not _.startswith("pymyinstall.") and
               not _.startswith("zip_safe flag not set; analyzing archive contents...") and
               not _.startswith("error removing build") and
               "UserWarning:" not in _ and
               "warnings.warn(" not in _ and
               "module references __file__" not in _]
        err = "\n".join(_ for _ in err if _)

        exp = ".zip/lib2to3/Grammar.txt"
        if len(err) > 0 and exp not in out.replace("\\", "/").replace("//", "/"):
            raise Exception(
                "Something went wrong:\nCMD\n{0}\nOUT\n{1}\nERR-C\n{2}".format(
                    cmd, out, err_keep))
        fLOG(out)

        dirpyexe = os.path.dirname(pyexe)
        fLOG(
            "[install_python] add python to PATH='{0}'".format(dirpyexe))
        path = os.environ['PATH']
        path = ";".join([dirpyexe, path])
        os.environ['PATH'] = path

        fLOG("[install_python] install modules")
        pattern = ('"{0}" -u -c "import sys;from pymyinstall.packaged import install_all;install_all'
                   '(fLOG=print, temp_folder=\'{2}\','
                   'verbose=True, source=\'2\', list_module=\'{1}\')"')
        cmd = pattern.format(
            pyexe, modules, download_folder.replace("\\", "/"))
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                           communicate=False, catch_exit=True)
        cmds.append(cmd)
        fLOG("[install_python] end installed modules.")
        if len(err) > 0:
            # We try a second time to make sure a second pass does not help.
            fLOG("[install_python2] install modules")
            out_, err_ = run_cmd(
                cmd, wait=True, fLOG=fLOG, communicate=False, catch_exit=False)
            err__ = _clean_err0(err_)
            if len(err__) > 0:
                mes = "[install_python2] end installed modules. Something went wrong:\n"
                raise Exception(
                    mes + "ERR-D-CMD\n{0}\nOUT\n{1}\nOUT2\n{3}\nERR-D\n{2}\nERR2-D\n{4}\nERR2-Dc\n{5}\n**CMD**\n{0}\n--CMDS--\n{6}".format(
                        cmd, out, err, out_, err_, err__, "\n".join(cmds)))
            out += ("\n-------------" * 5) + "\n" + out_
            fLOG("[install_python2] end installed modules.")
        fLOG(out)

    return local


def folder_older_than(folder, delay=datetime.timedelta(30)):
    """
    Tells if a folder is older than a given timespan.

    @param      folder      folder name
    @param      delay       delay
    @return                 boolean
    """
    folder = os.path.abspath(folder)
    if not os.path.exists(folder):
        return False
    cre = os.stat(folder).st_ctime
    dt = datetime.datetime.fromtimestamp(cre)
    now = datetime.datetime.now()
    delta = now - dt
    return delta > delay



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import ModuleInstall


class TestInstallModuleIsInstalled (unittest.TestCase):

    def test_module_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = ModuleInstall("autopep8", "pip", fLOG=fLOG)
        self.assertTrue(mod.is_installed_local())
        self.assertTrue(mod.is_installed_local_cmd())


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Helpers for pip.

.. versionadded:: 1.5
"""
try:
    from pip._internal.utils.compat import stdlib_pkgs
except ImportError:
    stdlib_pkgs = None


class Distribution:
    """
    Common interface for old and recent pip packages.

    .. versionadded:: 1.5
    """

    def __init__(self, dist):
        self.dist = dist

    def __getattr__(self, attr):
        if attr == 'key':
            if hasattr(self.__dict__['dist'], 'key'):
                return self.__dict__['dist'].key
            return self.__dict__['dist'].canonical_name
        if attr == 'dist':
            return self.__dict__['dist']
        if attr in {'_get_metadata', 'requires', 'PKG_INFO', 'project_name'}:
            if hasattr(self.__dict__['dist'], attr):
                return getattr(self.__dict__['dist'], attr)
            return getattr(self.__dict__['dist']._dist, attr)
        return getattr(self.__dict__['dist'], attr)


def get_installed_distributions(local_only=True, skip=stdlib_pkgs,
                                include_editables=True, editables_only=False,
                                user_only=False, use_cmd=False):
    """
    Directs call to function *get_installed_distributions* from :epkg:`pip`.

    Return a list of installed Distribution objects.

    :param local_only: if True (default), only return installations
        local to the current virtualenv, if in a virtualenv.
    :param skip: argument is an iterable of lower-case project names to
        ignore; defaults to ``pip.compat.stdlib_pkgs`` (if *skip* is None)
    :param editables: if False, don't report editables.
    :param editables_only: if True , only report editables.
    :param user_only: if True , only report installations in the user
        site directory.
    :param use_cmd: if True, use a different process (updated package list)
    :return: list of installed Distribution objects.

    .. versionadded:: 1.5
    """
    if use_cmd:
        raise NotImplementedError("use_cmd should be False")
    try:
        from pip._internal.metadata import get_default_environment
        return list(map(Distribution,
                        get_default_environment().iter_installed_distributions(
                            local_only=local_only, skip=skip,
                            include_editables=include_editables,
                            editables_only=editables_only,
                            user_only=user_only)))

    except ImportError:
        from pip._internal.utils.misc import get_installed_distributions as getd
        return list(map(Distribution, getd(
            local_only=local_only, skip=skip,
            include_editables=include_editables,
            editables_only=editables_only,
            user_only=user_only, use_cmd=use_cmd)))



"""
@file
@brief Install or update all packages.

.. faqref::
    :title: Skipped modules

    Some modules are still marked as a dependency by others
    even if they are not needed or cannot be installed.

    * `enum-compat <https://pypi.python.org/pypi/enum-compat>`_:
      not needed after Python 3.5
    * `prettytable <https://pypi.python.org/pypi/PrettyTable>`_:
      should be repladed by `PTable <https://pypi.python.org/pypi/PTable>`_.
"""
from __future__ import print_function
import sys
import os
import warnings
import time
from textwrap import wrap
from ..installhelper import ModuleInstall, has_pip, update_pip, is_installed, get_module_dependencies
from ..installhelper.module_install_exceptions import MissingVersionOnPyPiException, MissingPackageOnPyPiException, MissingReferenceException
from ..installhelper.module_dependencies import missing_dependencies
from .packaged_exception import ModuleNotFoundError
from .packaged_config import small_set


def _build_reverse_index():
    """
    builds a reverse index of the module,
    """
    res = {}
    mods = small_set()
    for m in mods:
        res[m.name] = m
        if m.mname is not None and m.mname != m.name:
            res[m.mname] = m
        if m.name.lower() not in (m.name, m.mname):
            res[m.name.lower()] = m
    return res


_reverse_module_index = _build_reverse_index()


def find_module_install(name, must_exist=False):
    """
    Checks if there are specific instructions
    to run before installing module *name*,
    on Windows, some modules requires compilation,
    if not uses default option with *pip*.

    @param      name        module name, the name can include a specific version number with '=='
    @param      must_exist  if True, raise an exception if not found
    @return                 @see cl ModuleInstall
    """
    if isinstance(name, ModuleInstall):
        return name

    if name in {"pip", "python"}:
        return None

    if '=' in name:
        spl = name.split('==')
        if len(spl) != 2:
            raise ValueError(
                "[find_module_install] unable to interpret " + name)
        name = spl[0]
        version = spl[1]
    else:
        version = None

    if name in _reverse_module_index:
        mod = _reverse_module_index[name]
    else:
        names = [name[0].upper() + name[1:],
                 name.replace("-python", ""),
                 name + "-python"]
        if name == "pytables":
            names.append("tables")
        elif name == "pyopengl-accelerate":
            names.append("PyOpenGL_accelerate")
        elif name == "pywin32":
            names.append("pywin32")  # pypiwin32
        for n in names:
            if n in _reverse_module_index:
                return _reverse_module_index[n]
        if must_exist:
            raise MissingReferenceException(
                "Unable to find reference for module '{0}'\nCheck '{0}'".format(name))
        mod = ModuleInstall(name, "pip")

    if version is not None:
        mod.version = version
    return mod


def reorder_module_list(list_module):
    """
    reorder a list of modules to install, a module at position *p*
    should not depend not module at position *p+1*

    @param      list_module     list of module (@see cl ModuleInstall)
    @return                     list of modules

    The function uses modules stored in :mod:`pyminstall.packaged.packaged_config`,
    it does not go to pypi. If a module was not registered, it will be placed
    at the end in the order it was given to this function.
    """
    inset = {m.name: m for m in list_module}
    res = []
    for mod in small_set():
        if mod.name in inset and inset[mod.name] is not None:
            res.append(mod.copy(version=inset[mod.name].version))
            inset[mod.name] = None
    for mod in list_module:
        if inset[mod.name] is not None:
            res.append(mod)
    return res


class FileShouldNotBeFound(Exception):
    """
    Raised by function @see fn check_sys_path.
    """
    pass


def check_sys_path():
    """
    An issue is happening during unit test as pymyinstall
    could be imported from two locations. We check this is
    not the case.
    """
    exp = "pymyinstall_ut_skip_pyquickhelper_%d%d_std" % sys.version_info[:2]
    exp = os.path.join(exp, "src")
    for path in sys.path:
        if exp in path:
            raise FileShouldNotBeFound(
                "'{0}' was found in sys.path:\n{1}\n----\ncwd='{2}'\nexe='{3}'".format(path, "\n".join(sys.path), os.getcwd(), sys.executable))


def update_all(temp_folder=".", fLOG=print, verbose=True,
               list_module=None, reorder=True,
               skip_module=None, schedule_only=False,
               source=None, download_only=False):
    """
    Updates modules in *list_module*
    if None, this list will be returned by @see fn ensae_fullset,
    the function starts by updating :epkg:`pip`.

    @param      temp_folder     temporary folder
    @param      verbose         more display
    @param      list_module     None or of list of str or @see cl ModuleInstall or a name set
    @param      fLOG            logging function
    @param      reorder         reorder the modules to update first modules with less dependencies (as much as as possible)
    @param      skip_module     module to skip (list of str)
    @param      schedule_only   if True, the function returns the list of modules scheduled to be installed
    @param      source          overwrite the wheels location, see @see me get_exewheel_url_link2
    @param      download_only   only downloads the module, no installation

    *list_module* can be a set of modules of a name set.
    Name sets can be accesses by function @see fn get_package_set.
    See the function to get the list of possible name sets.

    .. versionchanged:: 1.1
        Catch an exception while updating modules and walk through the end of the list.
        The function should be run a second time to make sure an exception remains.
        It can be due to python keeping in memory an updated module.
        Parameters *source*, *download_only* were added.
    """
    check_sys_path()
    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)
    if not has_pip():
        fLOG("install pip")
        from .get_pip import main  # pylint: disable=E0401
        main()

    if skip_module is None:
        skip_module = []

    if list_module is None:
        from ..packaged import small_set
        list_module = small_set()
    elif isinstance(list_module, str):
        from .packaged_config import get_package_set
        f = get_package_set(list_module)
        list_module = f()
    else:
        list_module = [find_module_install(mod) if isinstance(
            mod, str) else mod for mod in list_module]

    list_module = [_ for _ in list_module if _.name not in skip_module]
    check_sys_path()

    if reorder:
        list_module = reorder_module_list(list_module)

    if verbose:
        fLOG("update pip if needed")

    if "pip" not in skip_module:
        update_pip()

    modules = list_module
    again = []
    errors = []
    schedule = []
    for mod in modules:
        check_sys_path()
        is_installed = mod.is_installed_version()
        if not is_installed:
            continue
        if verbose:
            fLOG("[wait] 0.5 seconds")
        time.sleep(0.5)
        if verbose:
            fLOG("[update-check] ##", mod.name, "## [begin]")

        try:
            has_update = mod.has_update()
        except (MissingVersionOnPyPiException, MissingPackageOnPyPiException) as e:
            # this happens for custom made version such as xgboost
            fLOG("    - unable to check updates", e)
            has_update = False
        if not has_update:
            if verbose:
                fLOG("[update-check] ##", mod.name, "## [end no update]")
            continue

        ver = mod.get_pypi_version()
        inst = mod.get_installed_version()

        schedule.append(mod)
        if not schedule_only:
            m = "    - updating module  {0} --- {1} --> {2} (kind={3})" \
                .format(mod.name, inst, ver, mod.kind)
            fLOG(m)
            try:
                if download_only:
                    b = mod.download(temp_folder=temp_folder, source=source)
                else:
                    b = mod.update(temp_folder=temp_folder,
                                   log=verbose, source=source)
            except (SystemExit, Exception) as e:
                b = False
                m = "    - failed to update module  {0} --- {1} --> {2} (kind={3}) due to {4} ({5})" \
                    .format(mod.name, inst, ver, mod.kind, str(e), type(e))
                fLOG(m)
                errors.append((mod, e))
            if b:
                again.append(m)

        if verbose:
            fLOG("[update-check] ##", mod.name, "## [end]")

    if schedule_only:
        return schedule

    if verbose:
        fLOG("")
        fLOG("updated modules")
        for m in again:
            fLOG("  ", m)
        if len(errors) > 0:
            fLOG("failed modules")
            for m in errors:
                fLOG("  ", m[0], m[1])
    return None


def install_all(temp_folder=".", fLOG=print, verbose=True,
                list_module=None, reorder=True, skip_module=None,
                up_pip=True, skip_missing=False, deps=False,
                schedule_only=False, deep_deps=False,
                _memory=None, source=None, download_only=False,
                force=False):
    """
    Installs modules in *list_module*
    if None, this list will be returned by @see fn ensae_fullset,
    the function starts by updating pip.

    @param      temp_folder     temporary folder
    @param      verbose         more display
    @param      list_module     None or of list of str or @see cl ModuleInstall or a name set
    @param      fLOG            logging function
    @param      reorder         reorder the modules to update first modules with less dependencies (as much as as possible)
    @param      skip_module     module to skip (list of str)
    @param      up_pip          upgrade pip (pip must not be in *skip_module*)
    @param      skip_missing    skip the checking of the missing dependencies
    @param      deps            install the dependencies of the installed modules
    @param      schedule_only   if True, the function returns the list of modules scheduled to be installed
    @param      deep_deps       check dependencies for dependencies
    @param      _memory         stores installed packages, avoid going into an infinite loop
    @param      source          overwrite the location of the wheels, see @see me get_exewheel_url_link2
    @param      download_only   only downloads the module, no installation
    @param      force           force the installation or the download
    @return                     output streams

    *list_module* can be a set of modules of a name set.
    Name sets can be accesses by function @see fn get_package_set.
    See the function to get the list of possible name sets.

    About *name set*, the function can install a set of modules but these set of modules have dependencies:

    * @see fn extended_set, @see fn sphinx_theme_set requires @see fn small_set
    * @see fn ml_set, @see fn ensae_set, @see fn teachings_set, @see fn iot_set, @see fn scraping_set
      requires @see fn small_set and @see fn extended_set

    If you want to install a specific module with its dependencies, I suggest
    to use option *deep_deps*.

    .. versionadded:: 1.1
    """
    check_sys_path()
    fLOG("[install_all] begin, exe:", sys.executable)
    if _memory is None:
        _memory = {}
    deps = deps or deep_deps
    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)
    if not has_pip():
        fLOG("install pip")
        from ..installhelper.get_pip import main
        main()

    if skip_module is None:
        skip_module = []

    if list_module is None:
        list_module = small_set()
    elif isinstance(list_module, str  # unicode#
                    ):
        from .packaged_config import get_package_set
        f = get_package_set(list_module)
        list_module = f()
    else:
        list_module = [find_module_install(mod) if isinstance(
            mod, str) else mod for mod in list_module]

    list_module = [_ for _ in list_module if _.name not in skip_module]
    check_sys_path()

    if reorder:
        list_module = reorder_module_list(list_module)

    if up_pip and "pip" not in skip_module:
        if verbose:
            fLOG("update pip if needed")
        update_pip()

    # More information.
    sorted_names = ", ".join(sorted(_.name.lower() for _ in list_module))
    ordered_names = ", ".join(_.name for _ in list_module)
    fLOG('[install_all] sorted modules ----------------------------')
    fLOG("\n".join(wrap(sorted_names, width=100,
                        initial_indent='    ', subsequent_indent='    ')))
    fLOG('[install_all] modules installation order ----------------')
    fLOG("\n".join(wrap(ordered_names, width=100,
                        initial_indent='    ', subsequent_indent='    ')))
    fLOG('[install_all] begin ----------------------')

    modules = list_module
    again = []
    errors = []
    installed = []
    schedule = []
    out_streams_module = {}
    for mod in modules:
        check_sys_path()
        if mod.name in _memory:
            # already done
            continue
        if verbose:
            fLOG("[install-check] ##", mod.name, "## [begin]")
        if force or not mod.is_installed_version():
            schedule.append(mod)
            if not schedule_only:
                ver = mod.version
                m = "    - installing module '{0}' --- --> '{1}' (kind='{2}')" \
                    .format(mod.name, ver, mod.kind)
                fLOG(m)
                if deps:
                    fLOG("[install-check-dep] ##", mod.name, "## [begin]")
                    install_module_deps(mod.name, temp_folder=temp_folder,
                                        fLOG=fLOG, verbose=verbose, deps=deps, deep_deps=deep_deps,
                                        _memory=_memory, source=source, download_only=download_only,
                                        force=force)
                    fLOG("[install-check-dep] ##", mod.name, "## [end]")
                else:
                    out_streams = []
                    try:
                        if download_only:
                            b = mod.download(
                                temp_folder=temp_folder, source=source)
                        else:
                            b = mod.install(temp_folder=temp_folder,
                                            log=verbose, source=source,
                                            out_streams=out_streams)
                    except (SystemExit, Exception) as e:
                        b = False
                        m = "    - failed to update module '{0}' --- '{1}' --> '{2}' (kind='{3}') due to '{4}' ('{5}')" \
                            .format(mod.name, '', ver, mod.kind, str(e), type(e))
                        fLOG(m)
                        errors.append((mod, e))
                    if b:
                        again.append(m)
                        installed.append(mod)
                    if verbose:
                        ne = 0
                        for lll in out_streams:
                            if lll[-1]:
                                ne += 1
                        if ne > 0:
                            fLOG("###################")
                            fLOG("##", mod.name, "##")
                            for jj in out_streams:
                                for la, va in zip(("CMD", "OUT", "ERR-1"), jj):
                                    fLOG(la, va)
                            fLOG("###################.")
                    out_streams_module[mod.name] = out_streams
        if verbose:
            fLOG("[install-check] ##", mod.name, "## [end]")
        _memory[mod.name] = True

    if schedule_only:
        return schedule

    if verbose:
        fLOG("")
        fLOG("[install_all]" +
             (" downloaded modules" if download_only else "installed modules"))
        for m in again:
            fLOG("  ", m)
        fLOG("[install_all] end: ", len(errors), "errors")
        if len(errors) > 0:
            fLOG("[install_all] failed modules")
            for m in errors:
                fLOG("  ", m[0], m[1])
            fLOG("[install_all] end failed modules")

    if not skip_missing:
        fLOG("[install_all] check dependencies")
        miss = missing_dependencies()
        if len(miss) > 0:
            mes = "\n".join("Module '{0}' misses '{1}'".format(k, ", ".join(v))
                            for k, v in sorted(miss.items()))
            warnings.warn("[install_all] missing dependencies\n" + mes)
        fLOG("[install_all] end check dependencies")

    for k, v in sorted(out_streams_module.items()):
        fLOG("[%s] #########################" % k)
        if isinstance(v, list):
            for _ in v:
                for la, va in zip(("CMD", "OUT", "ERR-2"), _):
                    if va is not None and len(va) > 0:
                        fLOG(la, va)
        else:
            for la, va in zip(("CMD", "OUT", "ERR-3"), v):
                if va is not None and len(va) > 0:
                    fLOG(la, va)
        fLOG(".")
    return out_streams_module


def install_module_deps(name, temp_folder=".", fLOG=print, verbose=True, deps=True,
                        deep_deps=False, _memory=None, source=None,
                        download_only=False, force=False):
    """
    Installs a module with its dependencies,
    if a module is already installed, it installs
    the missing dependencies.

    @param      module              module name
    @param      temp_folder         where to download
    @param      fLOG                logging function
    @param      verbose             more display
    @param      deps                add dependencies
    @param      deep_deps           check dependencies for dependencies
    @param      _memory             stores installed packages, avoid going into an infinite loop
    @param      source              overwrite the wheels location, see @see me get_exewheel_url_link2
    @param      download_only       only downloads the module, no installation
    @param      force               force the download or the update
    @return                         list of installed modules

    The function does not handle properly contraints on versions.
    It checks in the registered list of modules if *name* is present.
    If it is the case, it uses it, otherwise, it uses *pip*.

    .. versionadded:: 1.1
    """
    check_sys_path()
    if _memory is None:
        _memory = {}
    deps = deps or deep_deps
    installed = []
    first_name = name
    memory2 = {}
    if not is_installed(name):
        install_all(temp_folder=temp_folder, fLOG=fLOG, verbose=verbose,
                    list_module=[name], reorder=True, up_pip=False,
                    skip_missing=True, _memory=memory2, deps=False,
                    deep_deps=False, source=source, download_only=download_only,
                    force=force)
        installed.append(name)

    stack = [name]
    while len(stack) > 0:
        check_sys_path()
        name = stack[-1]
        del stack[-1]
        if name in _memory:
            # already done, avoid loops on
            continue
        if name == first_name:
            _memory.update(memory2)

        res = get_module_dependencies(name, refresh_cache=True, use_pip=True)
        if res is None:
            raise ImportError("unable to check dependencies of module {0}\ninstalled:\n{1}".format(
                name, "\n".join(installed)))
        list_m = [k for k, v in res.items()]
        fLOG("[deps] [{}] requires".format(name), list_m)
        inst = [k for k in list_m if deep_deps or not is_installed(k)]
        fLOG("[deps] installation of ", inst)
        install_all(temp_folder=temp_folder, fLOG=fLOG, verbose=verbose,
                    list_module=inst, reorder=True, up_pip=False,
                    skip_missing=True, deep_deps=deep_deps,
                    _memory=_memory, source=source, download_only=download_only)
        stack.extend(inst)
        installed.extend(inst)
        for i in inst:
            _memory[i] = True

    return list(sorted(set(installed)))


def install_module(module_name, temp_folder=".", fLOG=print, verbose=True,
                   reorder=True, skip_module=None,
                   up_pip=True, skip_missing=False, deps=False,
                   schedule_only=False, deep_deps=False,
                   _memory=None, source=None, download_only=False,
                   force=False):
    """
    Installs a module.

    @param      module_name     module to install
    @param      temp_folder     temporary folder
    @param      verbose         more display
    @param      fLOG            logging function
    @param      reorder         reorder the modules to update first modules with less dependencies (as much as as possible)
    @param      skip_module     module to skip (list of str)
    @param      up_pip          upgrade pip (pip must not be in *skip_module*)
    @param      skip_missing    skip the checking of the missing dependencies
    @param      deps            install the dependencies of the installed modules
    @param      schedule_only   if True, the function returns the list of modules scheduled to be installed
    @param      deep_deps       check dependencies for dependencies
    @param      _memory         stores installed packages, avoid going into an infinite loop
    @param      source          overwrite the wheels location, see @see me get_exewheel_url_link2
    @param      download_only   only downloads the module, no installation
    @param      force           force the download or the install

    .. versionadded:: 1.1
    """
    if not isinstance(module_name, list):
        module_name = [module_name]
    install_all(list_module=module_name, temp_folder=temp_folder, fLOG=fLOG,
                verbose=verbose, reorder=reorder, skip_module=skip_module,
                up_pip=up_pip, skip_missing=skip_missing, deps=deps,
                schedule_only=schedule_only, deep_deps=deep_deps,
                _memory=_memory, source=source, download_only=download_only,
                force=force)


def update_module(module_name, temp_folder=".", fLOG=print, verbose=True,
                  reorder=True, skip_module=None, schedule_only=False,
                  source=None):
    """
    Updates modules in *list_module*
    if None, this list will be returned by @see fn ensae_fullset,
    the function starts by updating pip.

    @param      module_name     module to update
    @param      temp_folder     temporary folder
    @param      verbose         more display
    @param      fLOG            logging function
    @param      reorder         reorder the modules to update first modules with less dependencies (as much as as possible)
    @param      skip_module     module to skip (list of str)
    @param      schedule_only   if True, the function returns the list of modules scheduled to be installed
    @param      source          overwrite the wheels location, see @see me get_exewheel_url_link2
    """
    if not isinstance(module_name, list):
        module_name = [module_name]
    update_all(list_module=module_name, temp_folder=temp_folder, fLOG=fLOG, verbose=verbose,
               reorder=reorder, skip_module=skip_module, schedule_only=schedule_only,
               source=source)


def download_module(module_name,
                    temp_folder=".", force=False,
                    unzipFile=True, file_save=None, deps=False,
                    source=None, fLOG=print):
    """
    Downloads the module without installation.

    @param      temp_folder     destination
    @param      force           force the installation even if already installed
    @param      unzipFile       if it can be unzipped, it will be (for github, mostly)
    @param      file_save       for debug purposes, do not change it unless you know what you are doing
    @param      deps            download the dependencies too (only available for pip)
    @param      source          overwrite the wheels location, see @see me get_exewheel_url_link2
    @return                     downloaded files
    """
    if not isinstance(module_name, list):
        module_name = [module_name]

    res = []
    for mod in module_name:
        if not isinstance(mod, ModuleInstall):
            mod = find_module_install(mod, True)
            if mod is None:
                raise ModuleNotFoundError(
                    "unable to find an objec for module: " + ", ".join(module_name))
        k = mod.fLOG
        mod.fLOG = fLOG
        f = mod.download(temp_folder=temp_folder, force=force,
                         unzipFile=unzipFile, file_save=file_save, deps=deps, source=source)
        mod.fLOG = k
        res.append(f)
    return res



# -*- coding: utf-8 -*-
"""
@file
@brief Functions to help creating a setup

.. versionadded:: 1.1
"""
import os
import sys
from ..installhelper import run_cmd

_setup_py = """
# -*- coding: utf-8 -*-
import sys
import os
import warnings
from distutils.core import setup
from setuptools import find_packages

#########
# settings
#########


project_var_name = "__NAME__"
sversion = "__VERSION__"
subversion = "__SUBVERSION__"
versionPython = "%s.%s" % (sys.version_info.major, sys.version_info.minor)
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'


KEYWORDS = project_var_name
DESCRIPTION = '''__DESCRIPTION___'''
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 5 - Production/Stable'
]


#######
# data
#######

packages = find_packages('src', exclude='src')
package_dir = {k: "src/" + k.replace(".", "/") for k in packages}
package_data = {
    project_var_name : ["*.pyd", "*.dll", "*.so" ],
}


def verbose():
    print("---------------------------------")
    print("package_dir =", package_dir)
    print("packages    =", packages)
    print("package_data=", package_data)
    print("current     =", os.path.abspath(os.getcwd()))
    print("---------------------------------")


setup(
    name=project_var_name,
    version='%s%s' % (sversion, subversion),
    author='__AUTHOR__',
    author_email='__EMAIL__',
    url="__URL__",
    download_url="__DURL__",
    description=DESCRIPTION,
    long_description=long_description,
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    install_requires=[]
)
"""

_readme_rst = '''
.. _l-README:

README / Changes
================

.. image:: https://travis-ci.com/sdpython/pymyinstall.svg?branch=master
    :target: https://app.travis-ci.com/github/sdpython/pymyinstall
    :alt: Build status

.. image:: https://badge.fury.io/py/pymyinstall.svg
    :target: http://badge.fury.io/py/pymyinstall

.. image:: http://img.shields.io/pypi/dm/pymyinstall.png
    :alt: PYPI Package
    :target: https://pypi.python.org/pypi/pymyinstall

.. image:: http://img.shields.io/github/issues/sdpython/pymyinstall.png
    :alt: GitHub Issues
    :target: https://github.com/sdpython/pymyinstall/issues

.. image:: https://img.shields.io/badge/license-MIT-blue.svg
    :alt: MIT License
    :target: http://opensource.org/licenses/MIT


**Links:**

* `GitHub/__NAME__ <__URL__>`_
* `documentation <__URL__>`_
* `Blog <http:/__NAME__/helpsphinx/blog/main_0000.html#ap-main-0>`_
'''

if sys.version_info[0] == 2:
    from codecs import open


def create_empty_folder_setup(fold, name, author=None, description=None, url=None, durl=None, version="0.1",
                              subversion="0"):
    """
    Creates a quick empty shell for a new project.

    @param      fold        location
    @param      name        name of the project
    @param      author      author
    @param      description description
    @param      url         url for the documentation
    @param      durl        url to download it
    @param      version     version
    @param      subversion  third part of the version number
    @return                 list of created files
    """
    src = os.path.join(fold, "src")
    mod = os.path.join(src, name)
    create = [fold, src, mod]
    for f in create:
        if not os.path.exists(f):
            os.makedirs(f)

    replace = dict(__AUTHOR__=author,
                   __DESCRIPTION__=description,
                   __URL__=url,
                   __DURL__=durl,
                   __VERSION__=version,
                   __SUBVERSION__=subversion,
                   __NAME__=name)

    def adapt(content):
        for k, v in replace.items():
            if v is not None:
                content = content.replace(k, v)
        return content

    readme = os.path.join(fold, "README.rst")
    with open(readme, "w", encoding="utf-8") as f:
        f.write(adapt(_readme_rst))

    setup = os.path.join(fold, "setup.py")
    with open(setup, "w", encoding="utf-8") as f:
        f.write(adapt(_setup_py))

    writ = [setup, readme]

    if sys.platform.startswith("win"):
        bat = os.path.join(fold, "wheel.bat")
        with open(bat, "w") as f:
            f.write("{0} setup.py bdist_wheel".format(sys.executable))
        writ.append(bat)

    return writ


_setup_py_existing = """
# -*- coding: utf-8 -*-
import sys
import os
import warnings
from distutils.core import setup
from setuptools import find_packages

DESCRIPTION = '''__DESCRIPTION__'''

packages = find_packages('src', exclude='src')
package_dir = {k: 'src/' + k.replace(".", "/") for k in packages}
package_data = {
__PACKDATA__
}

setup(name="__FOLDER__", version="__VERSION__",
    description=DESCRIPTION, long_description=DESCRIPTION,
    packages=packages, package_dir=package_dir,
    package_data=package_data
)
"""


def create_folder_setup(fold, wheel=True, output_path=None, fLOG=None, version=None, description=None):
    """
    Creates a quick setup for an existing or installed projects.

    @param      fold            folder or module (must be imported first)
    @param      wheel           run the setup to build the wheel
    @param      output_path     copies everything here it not None
    @param      version         to overwrite version
    @param      description     to overwrite description
    @return                     list of created files

    .. exref::
        :title: Create a setup from installed packages
        :lid: ex-torch-setup

        This packages :epkg:`pandas` into a wheel from the installed
        sources. Location of the sources can be specified too.

        ::

            from pymyinstall.setuphelper import create_folder_setup
            create_folder_setup('pandas', fLOG=print, output_path='.')
    """
    if fLOG:
        fLOG("[create_folder_setup] process '{0}'".format(fold))
    if not os.path.exists(fold):
        if fold not in sys.modules:
            raise ValueError("Unable to find module '{0}'".format(fold))
        mod = sys.modules[fold]
        version = mod.__version__ if hasattr(
            mod, '__version__') else mod.VERSION
        info = dict(__VERSION__=version,
                    __DESCRIPTION__=mod.__doc__,
                    __FOLDER__=fold)
        fold = os.path.dirname(mod.__file__)
    else:
        info = {}

    # Module name
    name = os.path.split(fold)[-1]
    if fLOG:
        fLOG("[create_folder_setup] name='{0}'".format(name))
        fLOG("[create_folder_setup] fold='{0}'".format(fold))

    if output_path is not None:
        # Copies everything.
        if fLOG:
            fLOG("[create_folder_setup] copy to '{0}'".format(output_path))
        from pyquickhelper.filehelper import synchronize_folder
        dest = os.path.join(output_path, 'src', name)
        if not os.path.exists(dest):
            os.makedirs(dest)
        # , filter_copy=lambda name: '__pycache__' not in name)
        synchronize_folder(fold, dest)
        fold = dest
        fold_ = output_path
    else:
        raise ValueError("output_path must be specified")

    if len(info) == 0 and version is None:
        # Import the module.
        if name in sys.modules:
            del sys.modules[name]
        f = os.path.normpath(os.path.abspath(fold))
        f_ = os.path.normpath(os.path.join(fold, '..'))
        if fold_ not in sys.path:
            sys.path.insert(0, f_)
            rem = True
        else:
            rem = False

        mod = __import__(name)
        if rem:
            ind = sys.path.index(f_)
            del sys.path[ind]
        del sys.modules[name]

        info = dict(__VERSION__=mod.__version__,
                    __DESCRIPTION__=description or mod.__doc__,
                    __FOLDER__=name)
    else:
        info = dict(__VERSION__=version,
                    __DESCRIPTION__=description,
                    __FOLDER__=name)

    # Package data
    from pyquickhelper.filehelper import explore_folder
    _, files = explore_folder(fold, fullname=False)
    pack = {}
    for f in files:
        ff = os.path.split(f)[0]
        f = os.path.relpath(f, fold)
        if '__pycache__' in f:
            continue
        ext = os.path.splitext(f)[-1]
        if ext in {'.py', '.pyc'}:  # , '.pyd'}:
            continue
        f, n = os.path.split(f)
        f = f .replace("\\", "/").replace("/", ".")
        if f not in pack:
            pack[f] = []
            init = os.path.join(ff, "__init__.py")
            while not os.path.exists(init):
                if fLOG:
                    fLOG("[create_folder_setup] add '{0}'".format(init))
                with open(init, "w") as fi:
                    fi.write('# added for additional files')
                ff = os.path.split(ff)[0]
                init = os.path.join(ff, "__init__.py")
        pack[f].append(n)

        if fLOG:
            fLOG("[create_folder_setup] add '{0}' in '{1}'".format(n, f))

    rows = ["  '{2}{3}{0}': {1},".format(
        k, v, name, '.' if k else '') for k, v in pack.items()]
    info['__PACKDATA__'] = "\n".join(rows)

    # Writes setup.py
    script = _setup_py_existing
    for k, v in info.items():
        if v is not None:
            script = script.replace(k, v)
    setup = os.path.join(fold_, 'setup.py')
    if fLOG:
        fLOG("[create_folder_setup] write='{0}'".format(setup))
    with open(setup, "w") as f:
        f.write(script)

    if wheel:
        cmd = '"{0}" -u setup.py bdist_wheel'.format(
            sys.executable.replace("w.exe", ".exe"))
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG, change_path=fold_)
        if fLOG:
            fLOG('[create_folder_setup] OUT --------------\n' + out)
            fLOG('[create_folder_setup] ERR --------------\n' + err)
        return [setup]
    else:
        return [setup]



"""
@brief      test log(time=103s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.win_installer import import_every_module


class TestLONGSuccessfulImport3(unittest.TestCase):

    def test_long_import_every_module(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        nb = int("".join(_ for _ in os.path.split(
            __file__)[-1] if "0" <= _ <= "9"))

        res = import_every_module(
            sys.executable, None, fLOG=fLOG, start=50 * nb, end=50 * (nb + 1))
        nb = 0
        for r in res:
            if not r[0]:
                fLOG("---------------------------------")
                fLOG("FAILED", r[1], "\nOUT\n", r[2], "\nERR\n", r[3])
            else:
                nb += 1
        assert nb > 0


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=2s)
"""
import unittest
import pandas
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pandashelper import df2rst
from pymyinstall.packaged import get_package_set, name_sets_dataframe


class TestNameSet(unittest.TestCase):

    def test_documentation(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        r = name_sets_dataframe()
        for m in r:
            fLOG("**", m)
        self.assertGreater(len(r), 1)
        df = pandas.DataFrame(r)
        df = df[["name", "description"]]
        rst = df2rst(df)
        fLOG(rst)
        self.assertGreater(len(rst), 1)

        nb = 0
        for mod in r:
            lp = get_package_set(mod["name"])
            if len(lp()) == 0 and mod["name"] != "pywin32":
                raise Exception("issue with module '{0}'".format(mod["name"]))
            nb += 1
        self.assertGreater(nb, 1)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=3s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.win_installer.win_packages import get_modules_version


class TestPackages(unittest.TestCase):

    def test_modules_list(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mods = get_modules_version(os.path.dirname(sys.executable))
        assert len(mods) > 0
        lastv = ""
        for v in mods.values():
            lastv = v
            break
        self.assertIn(".", lastv)

    def test_modules_list_ext(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        path = "C:\\github\\ensae_teaching_cs\\dist\\win_python_setup\\python"
        if os.path.exists(path):
            mods = get_modules_version(path)
            assert len(mods) > 0


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.module_install import ModuleInstall, get_pypi_version
from pymyinstall.installhelper.module_install_exceptions import MissingInstalledPackageException


class TestUpdateModule (unittest.TestCase):

    def test_update_module(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_update_module")

        vers = get_pypi_version("pymyinstall", True)
        fLOG(vers)
        assert len(vers) > 0

        try:
            mod = ModuleInstall("joblib", "pip", fLOG=fLOG)
            mod.update(temp_folder=temp, source="2")
        except MissingInstalledPackageException:
            # not installed
            pass

        mod = ModuleInstall("pandas", "wheel", fLOG=fLOG)
        mod.update(temp_folder=temp, source="2")

        mod = ModuleInstall("xlrd", "pip", fLOG=fLOG)
        mod.update(temp_folder=temp, source="2")


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Shortcuts to tutorial
"""
import os


def copy_tutorial(name, destination):
    """
    copy files and scripts for a specific tutorial

    @param      name            tutorial name or folder
    @param      destination     destination
    @return                     list of operations done by the function list of 3-uple: action, source_file, dest_file

    The function will create a sub folder in *destination*
    using *name* or the last folder name in *name*.

    This function requires modules
    `pyquickhelper <http://www.xavierdupre.fr/app/pyquickhelper/helpsphinx/>`_.
    """
    from pyquickhelper.filehelper import synchronize_folder
    if os.path.exists(name):
        dest = os.path.join(destination, name)
    else:
        this = os.path.abspath(os.path.dirname(__file__))
        fold = os.path.join(this, name)
        if not os.path.exists(fold):
            raise FileNotFoundError(
                "unable to find tutorial {0}\n{1} not here".format(name, fold))
        spl = os.path.split(fold)
        dest = os.path.join(destination, spl[-1])

    if not os.path.exists(dest):
        os.mkdir(dest)

    return synchronize_folder(fold, dest)



# -*- coding: utf-8 -*-
"""
@file
@brief Run some checkings about the distribution

.. versionadded:: 1.1
"""
from __future__ import print_function

import os
import sys

from .win_exception import WinInstallDistributionError
from ..packaged.automate_install import find_module_install
from ..installhelper.install_venv_helper import run_cmd_path
from ..packaged import small_set
from ..installhelper.module_install import ModuleInstall, run_cmd

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def distribution_checkings(python_path, tools_path, fLOG=print, skip_import=False, module_list=None):
    """
    checks a distribution was properly executed

    @param      python_path     path for python
    @param      tools_path      path for tools
    @param      fLOG            logging function
    @param      skip_import     skip the validation of every installed module (for unit test purposes)
    @param      module_list     module list to check

    The function raises @see cl WinInstallDistributionError if an issue is detected.

    If *python_path* is None, it is replace by ``os.path.dirname(sys.executable)``.

    .. versionadded:: 1.1
    """
    exceptions = []
    files_to_check = ["jupyter-console.exe", "jupyter-qtconsole.exe",
                      "jupyter-notebook.exe", "jupyter.exe"]

    if python_path is None:
        python_path = os.path.dirname(sys.executable)

    # on Python 3.5, pip does not always exist
    # in a virtual environment (probable bug)
    pip = os.path.join(python_path, "pip.exe")
    pep8 = os.path.join(python_path, "pep8.exe")
    if not os.path.exists(pip) and not os.path.exists(pep8):
        scripts = os.path.join(python_path, "Scripts")
        files_to_check.extend(
            [("spyder.bat", 'spyder.exe'), "autopep8.exe"])
        if sys.version_info[:2] != (3, 5):
            files_to_check.append("pip.exe")
    else:
        scripts = python_path

    ######################################
    # check Jupyter, numpy, works properly
    ######################################
    if sys.platform.startswith("win"):
        new_files_to_check = []
        for file in files_to_check:
            if isinstance(file, tuple):
                fs = [os.path.join(scripts, f) for f in file]
                ft = [f for f in fs if os.path.exists(f)]
                if len(ft) > 0:
                    new_files_to_check.append(os.path.split(ft[0])[-1])
                else:
                    try:
                        raise FileNotFoundError(fs)
                    except Exception as e:
                        exceptions.append(e)
            else:
                f = os.path.join(scripts, file)
                if not os.path.exists(f):
                    try:
                        raise FileNotFoundError(f)
                    except Exception as e:
                        exceptions.append(e)
                else:
                    new_files_to_check.append(file)
        files_to_check = new_files_to_check

    ################################
    # check that module can be imported
    ################################
    if not skip_import:
        res = import_every_module(
            python_path, module_list, only_installed=True, fLOG=fLOG)
        mes = []
        for r in res:
            if not r[0]:
                m = "------\nFAILED {0}\nOUT\n{1}\nERR--E\n{2}".format(
                    r[1].name, r[2], r[3])
                mes.append(m)
        if len(mes) > 0:
            raise WinInstallDistributionError(
                "cannot import modules\n" + "\n".join(mes))

    ########
    # final
    ########
    if len(exceptions) > 0:
        typstr = str  # unicode#
        errors = "\n----\n".join("{0}\n{1}".format(type(_), typstr(_))
                                 for _ in exceptions)
        raise WinInstallDistributionError("\n" + errors)


def import_every_module(python_path, module_list, only_installed=True, fLOG=print, start=0, end=-1):
    """
    import every module in *module_list* assuming they are defined
    by @see cl ModuleInstall or a string

    @param      python_path     python path
    @param      module_list     module list, if None, consider the largest list
    @param      only_installed  True to check only installed module in the list, False to test them without checking they were installed
    @param      start           start the list at *start*
    @param      end             end the list at *end* or -1 for all
    @return                     list of tuple (success, failing modules, output, error)
    """
    if python_path is None:
        python_path = os.path.dirname(sys.executable)
    if os.path.isfile(python_path):
        python_path = os.path.dirname(python_path)

    if module_list is None:
        module_list = small_set()

    def tomod(m):
        if isinstance(m, ModuleInstall):
            return m
        else:
            return find_module_install(m)
    module_list = [tomod(m) for m in module_list]

    def noLOG(*l, **p):
        pass

    def is_errored_line(line):
        if ".py" in line:
            if "UserWarning:" in line:
                return False
            if "FutureWarning:" in line:
                return False
            return True
        return False

    def analyze_error_success(mod, err):
        if err is None or len(err) == 0:
            return True
        if "ImportError" in err:
            return False
        if "ShimWarning" in err:
            return True
        if "kivy" in mod.name.lower():
            if "error" not in err.lower():
                return True

        if err is not None:
            lines = err.split("\n")
            for line in lines:
                if ".py" in line:
                    res = is_errored_line(lines)
                    if res:
                        return False
                else:
                    continue
            return True
        return False

    res = []
    for i, m in enumerate(module_list):
        if i < start:
            continue
        if end != -1 and i >= end:
            fLOG("[pymy] {0}/{1}: end".format(i, len(module_list) - start))
            break
        if m.is_installed_version():

            if m.name in ["libpython", "tutormagic", "pymyinstall", "distribute"]:
                # nothing to import or failure
                fLOG("[pymy] {0}/{1}: skipped".format(i,
                                                      len(module_list) - start), m)
                continue
            elif m.mname == "theano":
                # we need to check that TDM-GCC is installed
                cmd = "g++ --help"
                out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
                if err is not None and len(err) > 0:
                    fLOG("[pymy] {0}/{1}: failed (g++)".format(i,
                                                               len(module_list) - start), m)
                    res.append((False, m, out, err))
                    continue
                if "Usage: g++ [options] file..." not in out:
                    fLOG("[pymy] {0}/{1}: failed (g++)".format(i,
                                                               len(module_list) - start), m)
                    res.append((False, m, out, err))
                    continue

            sc = "import " + m.ImportName
            if m.name == "cvxopt":
                # see
                # http://stackoverflow.com/questions/14778178/import-cvxopt-base-the-specified-module-could-not-be-found
                sc += "import numpy;" + sc
            if m.name == "scipy":
                # see
                # http://stackoverflow.com/questions/14778178/import-cvxopt-base-the-specified-module-could-not-be-found
                sc += sc + ";import scipy.stats"
            out, err = run_cmd_path(
                python_path, sc, fLOG=fLOG, communicate=True, timeout=40)
            suc = analyze_error_success(m, err)
            nextm = module_list[i + 1] if i + 1 < len(module_list) else ""
            if suc:
                fLOG("[pymy] {0}/{1}: success".format(i,
                                                      len(module_list) - start), m, "-->", nextm)
            else:
                fLOG("[pymy] {0}/{1}: failed ".format(i,
                                                      len(module_list) - start), m, "-->", nextm)
                if m.name == "paramiko":
                    err = "You might have to install manually pycrypto.\n" + \
                          "Please read http://www.xavierdupre.fr/app/pymyinstall/helpsphinx//blog/2016/2016-02-27_pycrypto_paramiko.html" + \
                          err
                err = [(" - ERR--F: " if is_errored_line(line)
                        else " - OK:  ") + line.rstrip("\n\r") for line in err.split("\n")]
                err = "\n".join(err)
            res.append((suc, m, out, err))
    return res



"""
@brief      test log(time=45s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installcustom import install_scite


class TestScite (unittest.TestCase):

    def test_install_scite(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.version_info[0] == 2:
            return

        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_scite")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            f = os.path.join(temp, _)
            if os.path.isfile(f):
                os.remove(f)

        tt = os.path.join(temp, "wscite")
        if os.path.exists(tt):
            for _ in os.listdir(tt):
                f = os.path.join(tt, _)
                if os.path.isfile(f):
                    os.remove(f)

        if sys.platform.startswith("win"):
            r = install_scite(temp, fLOG=fLOG, change_python_path=True)
            exe = os.path.abspath(r)
            self.assertTrue(os.path.exists(exe))
            conf = exe.replace("SciTE.exe", "python.properties")
            fLOG(conf)
            self.assertTrue(os.path.exists(conf))
            with open(conf, "r") as f:
                content = f.read().lower()
            if sys.executable.lower() not in content and \
               sys.executable.lower().replace(".exe", "w.exe") not in content:
                raise Exception(
                    "{0} or {1} not in \n{2}".format(
                        sys.executable,
                        sys.executable.replace(
                            ".exe",
                            "w.exe"),
                        content))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Exceptions raised during the installation of a module
"""


class MissingPackageOnPyPiException(Exception):

    """
    raised when a package is not found on pipy
    """
    pass


class MissingInstalledPackageException(Exception):

    """
    raised when a package is not installed
    """
    pass


class AnnoyingPackageException(Exception):

    """
    raised when a package is not on pypi
    """
    pass


class MissingVersionOnPyPiException(Exception):

    """
    raised when a version is missing on pipy
    """
    pass


class MissingVersionWheelException(Exception):

    """
    raised when a version is missing as a wheel
    """
    pass


class MissingWheelException(Exception):

    """
    raised when a wheel is missing
    """
    pass


class MissingReferenceException(Exception):

    """
    raised when a module is not referenced by this package
    """
    pass


class InstallError(Exception):
    """
    raised when a package cannot be installed
    """
    pass


class DownloadError(Exception):
    """
    raised when a package cannot be downloaded
    """
    pass


class ConfigurationError(Exception):
    """
    raised when something is wrong the current configuration
    """
    pass


class UpdatePipError(Exception):
    """
    raised when pip cannot be update or reinstalled
    """
    pass


class RunCmdError(Exception):
    """
    raised when a command line cannot be run
    """
    pass


class WrongVersionError(Exception):
    """
    cannot interpret a version
    """
    pass


class WrongWheelException(Exception):
    """
    raised when the downloaded wheel seems wrong
    """
    pass


class UnavailableCustomBuildError(Exception):
    """
    raise when a module does not have a custom build
    """
    pass



"""
@brief      test log(time=1s)
"""
import unittest
import pandas
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pandashelper import df2rst
from pymyinstall.installhelper.module_install_version import get_module_version, get_module_metadata
from pymyinstall.packaged import small_set


class TestInstallModuleVersion(unittest.TestCase):

    def test_module_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        res = get_module_version("matplotlib")
        self.assertTrue(len(str(res)) > 0)

        res = get_module_metadata("matplotlib")
        fLOG(res)
        self.assertTrue(isinstance(res, dict))

    def test_all_module_summary(self):

        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = small_set()
        mod.sort()
        df = pandas.DataFrame(_.as_dict(rst_link=True) for _ in mod)
        df = df[["usage", "rst_link", "kind", "version", "installed",
                 "license", "purpose", "classifier"]]
        df.columns = ["usage", "name", "kind", "version", "installed",
                      "license", "purpose", "classifier"]
        lic = df[~df.license.isnull()]
        # fLOG(lic[["name","license"]])
        fLOG("license", lic.shape)
        nolic = df[df.license.isnull()]
        fLOG("no license", nolic.shape)
        fLOG(nolic[["name", "license"]])
        self.assertTrue(lic.shape[0] > 0)

        rst = df2rst(df)
        # fLOG(rst)
        self.assertTrue(len(rst) > 1000)


if __name__ == "__main__":
    unittest.main()
